# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Convergence test for Eagle Strategy."""

from vizier import benchmarks
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import random
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy
from vizier._src.algorithms.designers.eagle_strategy import testing
from vizier._src.algorithms.testing import comparator_runner
from vizier._src.benchmarks.experimenters.synthetic import bbob

from absl.testing import absltest
from absl.testing import parameterized


class EagleStrategyConvergenceTest(parameterized.TestCase):
  """Convergence test for Eagle Strategy designer.

  Note that all optimization problems are MINIMIZATION.
  """

  @parameterized.parameters(
      testing.create_continuous_exptr(bbob.Gallagher101Me),
      testing.create_continuous_log_scale_exptr(bbob.Rastrigin),
      testing.create_categorical_exptr(),
  )
  def test_convergence(self, exptr):

    def _random_designer_factory(problem, seed):
      return random.RandomDesigner(problem.search_space, seed=seed)

    def _eagle_designer_factory(problem, seed):
      return eagle_strategy.EagleStrategyDesigner(problem, seed=seed)

    random_benchmark_state_factory = benchmarks.DesignerBenchmarkStateFactory(
        designer_factory=_random_designer_factory, experimenter=exptr)

    eagle_benchmark_state_factory = benchmarks.DesignerBenchmarkStateFactory(
        designer_factory=_eagle_designer_factory,
        experimenter=exptr,
    )
    evaluations = 1000
    # Random designer batch size is large to expedite run time.
    comparator_runner.SimpleRegretComparisonTester(
        baseline_num_trials=2 * evaluations,
        candidate_num_trials=evaluations,
        baseline_suggestion_batch_size=2 * evaluations,
        candidate_suggestion_batch_size=5,
        baseline_num_repeats=5,
        candidate_num_repeats=1,
        alpha=0.05,
        goal=vz.ObjectiveMetricGoal.MINIMIZE,
    ).assert_benchmark_state_better_simple_regret(
        baseline_benchmark_state_factory=random_benchmark_state_factory,
        candidate_benchmark_state_factory=eagle_benchmark_state_factory,
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for nasbench201_experimenter."""
from absl import logging
import nats_bench
from vizier import pyvizier
from vizier._src.algorithms.designers import random
from vizier._src.benchmarks.experimenters import nasbench201_experimenter
from absl.testing import absltest


class Nasbench201ExperimenterTest(absltest.TestCase):

  @absltest.skip("Files must be installed manually.")
  def test_experimenter(self):
    nasbench = nats_bench.NATStopology(
        file_path_or_dict=DEFAULT_NATS_TSS_DIR, fast_mode=True, verbose=False)
    experimenter = nasbench201_experimenter.NASBench201Experimenter(nasbench)
    problem_statement = experimenter.problem_statement()
    designer = random.RandomDesigner(
        search_space=problem_statement.search_space, seed=None)

    suggestions = designer.suggest(5)
    trials = [suggestion.to_trial() for suggestion in suggestions]
    experimenter.evaluate(trials)
    for trial in trials:
      logging.info('Evaluated Trial: %s', trial)
      self.assertEqual(trial.status, pyvizier.TrialStatus.COMPLETED)
      metric_name = problem_statement.metric_information.item().name
      eval_objective = trial.final_measurement_or_die.metrics[metric_name].value
      self.assertGreaterEqual(eval_objective, 0.0)
      self.assertLessEqual(eval_objective, 100.0)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for eagle strategy."""

from absl.testing import parameterized
import jax
from jax import numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier import pyvizier as vz
from vizier._src.algorithms.optimizers import eagle_strategy
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.jax import types
from vizier.pyvizier import converters
from vizier.pyvizier.converters import padding

from absl.testing import absltest

tfd = tfp.distributions


def _create_logits_vector_simple(
    categorical_features,
    categorical_features_batch,
    scale,
    categorical_sizes,
    max_categorical_size,
    config,
):
  n_batch = categorical_features_batch.shape[0]
  n_feat = categorical_features.shape[0]
  logits = np.zeros((n_batch, len(categorical_sizes), max_categorical_size))
  for i, s in enumerate(categorical_sizes):
    one_hot_features = np.zeros([n_feat, s])
    one_hot_features[np.arange(n_feat), categorical_features[:, i]] = 1

    one_hot_batch = np.zeros([n_batch, s])
    one_hot_batch[np.arange(n_batch), categorical_features_batch[:, i]] = 1

    features_change = np.matmul(
        scale, one_hot_features
    ) - one_hot_batch * np.sum(scale, axis=-1, keepdims=True)

    diff_category_logit = np.log(
        (1.0 - config.prob_same_category_without_perturbation) / (s - 1)
    )
    logits_i = np.zeros((n_batch, max_categorical_size)) + diff_category_logit
    logits_i[:, s:] = -np.inf
    logits_i[np.arange(n_batch), categorical_features_batch[:, i]] = np.log(
        config.prob_same_category_without_perturbation
    )
    logits_i[:, :s] = logits_i[:, :s] + features_change
    logits[:, i, :] = logits_i
  return logits


def _create_features_simple(
    features,
    rewards,
    features_batch,
    rewards_batch,
    config,
    n_features,
    categorical_sizes,
    max_categorical_size,
    seed,
):
  """A version of `_create_features` that materializes large intermediates."""
  # Only works with no parallel batch dimension.
  continuous_features_diffs = (
      features.continuous - features_batch.continuous[:, jnp.newaxis, :]
  )
  categorical_features_diffs = (
      features.categorical != features_batch.categorical[:, jnp.newaxis, :]
  )
  features_diffs = vb.VectorizedOptimizerInput(
      continuous=continuous_features_diffs,
      categorical=categorical_features_diffs,
  )
  dists = jax.tree_util.tree_map(
      lambda x: jnp.sum(jnp.square(x), axis=-1), features_diffs
  )
  directions = rewards - rewards_batch[:, jnp.newaxis]
  scaled_directions = jnp.where(
      directions >= 0.0, config.gravity, -config.negative_gravity
  )

  # Handle removed fireflies without updated rewards.
  finite_ind = jnp.isfinite(rewards).astype(directions.dtype)

  # Ignore fireflies that were removed from the pool.
  scale = jax.tree_util.tree_map(
      lambda x: finite_ind  # pylint: disable=g-long-lambda
      * scaled_directions
      * jnp.exp(-config.visibility * x / n_features * 10.0),
      dists,
  )

  # Separate forces to pull and push so to normalize them separately.
  new_continuous_features = features_batch.continuous + jnp.sum(
      features_diffs.continuous * scale.continuous[..., jnp.newaxis], axis=1
  )
  categorical_features_logits = _create_logits_vector_simple(
      features.categorical,
      features_batch.categorical,
      scale.categorical,
      categorical_sizes,
      max_categorical_size,
      config,
  )
  new_categorical_features = tfd.Categorical(
      logits=categorical_features_logits
  ).sample(seed=seed)

  return vb.VectorizedOptimizerInput(
      new_continuous_features, new_categorical_features
  )


class VectorizedEagleStrategyContinuousTest(parameterized.TestCase):

  def setUp(self):
    super(VectorizedEagleStrategyContinuousTest, self).setUp()
    self.config = eagle_strategy.EagleStrategyConfig(
        visibility=1, gravity=1, pool_size=4
    )
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    root.add_categorical_param('c1', ['a', 'b'])
    root.add_categorical_param('c2', ['a', 'b', 'c'])
    self.converter = converters.TrialToModelInputConverter.from_problem(problem)
    self.eagle = eagle_strategy.VectorizedEagleStrategyFactory(
        eagle_config=self.config
    )(converter=self.converter, suggestion_batch_size=2)

  def test_create_features_and_logits(self):
    features_continuous = jnp.array(
        [[[1.0, 2.0]], [[3.0, 4.0]], [[7.0, 7.0]], [[8.0, 8.0]]]
    )
    features_categorical = jnp.array([[1, 2], [0, 0], [0, 1], [1, 1]])[
        :, jnp.newaxis, :
    ]
    rewards = jnp.array([2, 3, 4, 1])
    seed = jax.random.PRNGKey(0)
    features_continuous_batch = features_continuous[: self.eagle.batch_size]
    features_categorical_batch = features_categorical[: self.eagle.batch_size]
    features = vb.VectorizedOptimizerInput(
        continuous=features_continuous, categorical=features_categorical
    )
    features_batch = vb.VectorizedOptimizerInput(
        continuous=features_continuous_batch,
        categorical=features_categorical_batch,
    )
    rewards_batch = rewards[: self.eagle.batch_size]
    created_features = self.eagle._create_features(
        features,
        rewards,
        features_batch,
        rewards_batch,
        vb.VectorizedOptimizerInput(
            jnp.zeros_like(features_continuous_batch),
            jnp.zeros(features_categorical_batch.shape + (3,)),
        ),
        seed=seed,
    )
    self.assertEqual(created_features.continuous.shape, (2, 1, 2))
    self.assertEqual(created_features.categorical.shape, (2, 1, 2))

    features_2d = vb.VectorizedOptimizerInput(
        features.continuous[:, 0, :], features.categorical[:, 0, :]
    )
    features_batch_2d = vb.VectorizedOptimizerInput(
        features_batch.continuous[:, 0, :], features_batch.categorical[:, 0, :]
    )
    expected = _create_features_simple(
        features_2d,
        rewards,
        features_batch_2d,
        rewards_batch,
        self.config.replace(
            mutate_normalization_type=(
                eagle_strategy.MutateNormalizationType.UNNORMALIZED
            )
        ),
        (
            self.eagle.n_feature_dimensions.continuous
            + self.eagle.n_feature_dimensions.categorical
        ),
        self.eagle.categorical_sizes,
        self.eagle.max_categorical_size,
        seed,
    )
    actual = self.eagle._create_features(
        features,
        rewards,
        features_batch,
        rewards_batch,
        vb.VectorizedOptimizerInput(
            jnp.zeros_like(features_continuous_batch),
            jnp.zeros(features_categorical_batch.shape + (3,)),
        ),
        seed=seed,
    )
    np.testing.assert_array_equal(
        expected.continuous, actual.continuous[:, 0, :]
    )
    np.testing.assert_array_equal(
        expected.categorical, actual.categorical[:, 0, :]
    )

    scale = np.random.normal(size=[self.eagle.batch_size, 4])
    expected_logits = _create_logits_vector_simple(
        features_2d.categorical,
        features_batch_2d.categorical,
        scale,
        self.eagle.categorical_sizes,
        self.eagle.max_categorical_size,
        self.config,
    )
    actual_logits = self.eagle._create_categorical_feature_logits(
        features.categorical, features_batch.categorical, scale
    )
    np.testing.assert_allclose(
        expected_logits, actual_logits[:, 0, :, :], rtol=1e-5
    )

  @parameterized.parameters(1, 5)
  def test_create_random_perturbations(self, n_parallel):
    seed = jax.random.PRNGKey(0)
    perturbations_batch = jnp.ones(self.eagle.batch_size)
    perturbations = self.eagle._create_random_perturbations(
        perturbations_batch,
        n_parallel=n_parallel,
        seed=seed,
    )
    self.assertEqual(perturbations.continuous.shape, (2, n_parallel, 2))
    self.assertEqual(perturbations.categorical.shape, (2, n_parallel, 2, 3))

  def test_update_pool_features_and_rewards(self):
    features = vb.VectorizedOptimizerInput(
        continuous=jnp.array(
            [[[1, 2]], [[3, 4]], [[7, 7]], [[8, 8]]], dtype=jnp.float64
        ),
        categorical=jnp.array(
            [[[1, 2]], [[3, 0]], [[0, 1]], [[2, 1]]], dtype=jnp.int32
        ),
    )
    rewards = jnp.array([2, 3, 4, 1], dtype=jnp.float64)
    perturbations = jnp.array([1, 1, 1, 1], dtype=jnp.float64)

    batch_features = vb.VectorizedOptimizerInput(
        continuous=jnp.array([[[9, 9]], [[10, 10]]], dtype=jnp.float64),
        categorical=jnp.array([[[0, 0]], [[1, 1]]], dtype=jnp.int32),
    )
    batch_rewards = jnp.array([5, 0.5], dtype=jnp.float64)

    new_features, new_rewards, new_perturbations = (
        self.eagle._update_pool_features_and_rewards(
            batch_features,
            batch_rewards,
            jax.tree_util.tree_map(
                lambda f: f[: self.eagle.batch_size], features
            ),
            rewards[: self.eagle.batch_size],
            perturbations[: self.eagle.batch_size],
        )
    )
    np.testing.assert_array_equal(
        new_features.continuous,
        np.array([[[9, 9]], [[3, 4]]], dtype=np.float64),
        err_msg='Features are not equal.',
    )
    np.testing.assert_array_equal(
        new_features.categorical,
        np.array([[[0, 0]], [[3, 0]]], dtype=np.int32),
        err_msg='Features are not equal.',
    )

    np.testing.assert_array_equal(
        new_rewards,
        np.array([5, 3], dtype=np.float64),
        err_msg='rewards are not equal.',
    )

    pc = self.config.penalize_factor
    np.testing.assert_array_almost_equal(
        new_perturbations,
        np.array([1, pc], dtype=np.float64),
        err_msg='Perturbations are not equal.',
    )

  def test_update_best_reward(self):
    # Test replacing the best reward.
    features = vb.VectorizedOptimizerInput(
        continuous=jnp.array(
            [[[1, 2]], [[3, 4]], [[7, 7]], [[8, 8]]], dtype=jnp.float64
        ),
        categorical=jnp.array(
            [[[1, 2]], [[3, 0]], [[0, 1]], [[2, 1]]], dtype=jnp.int32
        ),
    )
    rewards = jnp.array([2, 3, 4, 1], dtype=jnp.float64)
    state = eagle_strategy.VectorizedEagleStrategyState(
        iterations=jnp.array(0),
        features=features,
        rewards=rewards,
        best_reward=jnp.max(rewards),
        perturbations=jnp.ones_like(rewards),
    )
    batch_features = vb.VectorizedOptimizerInput(
        continuous=jnp.array([[[9, 9]], [[10, 10]]], dtype=jnp.float64),
        categorical=jnp.array([[[0, 0]], [[1, 1]]], dtype=jnp.int32),
    )
    batch_rewards = jnp.array([5, 0.5], dtype=jnp.float64)
    seed = jax.random.PRNGKey(0)
    new_state = self.eagle.update(seed, state, batch_features, batch_rewards)
    self.assertEqual(new_state.best_reward, 5.0)
    # Test not replacing the best reward.
    batch_rewards = jnp.array([2, 4], dtype=jnp.float64)
    new_new_state = self.eagle.update(
        seed, new_state, batch_features, batch_rewards
    )
    self.assertEqual(new_new_state.best_reward, 5.0)

  @parameterized.parameters(
      {'batch_size': 5, 'expected_batch_size': 5, 'max_pool_size': 50},
      {'batch_size': None, 'expected_batch_size': 50, 'max_pool_size': 50},
      {'batch_size': 5, 'expected_batch_size': 5, 'max_pool_size': 10},
      {'batch_size': None, 'expected_batch_size': 10, 'max_pool_size': 10},
  )
  def test_batch_size_and_pool_size(
      self, batch_size, expected_batch_size, max_pool_size
  ):
    problem = vz.ProblemStatement()
    root = problem.search_space.root
    for i in range(100):
      root.add_float_param(f'x{i}', 0.0, 1.0)
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    config = eagle_strategy.EagleStrategyConfig(max_pool_size=max_pool_size)
    eagle = eagle_strategy.VectorizedEagleStrategyFactory(eagle_config=config)(
        converter=converter, suggestion_batch_size=batch_size
    )
    self.assertEqual(eagle.pool_size, max_pool_size)
    self.assertEqual(eagle.batch_size, expected_batch_size)

  def test_trim_pool(self):
    pc = self.config.perturbation
    features_batch = vb.VectorizedOptimizerInput(
        continuous=jnp.array([[[1, 2]], [[3, 4]]], dtype=jnp.float64),
        categorical=jnp.array([[[1, 2]], [[3, 0]]], dtype=jnp.int32),
    )
    rewards_batch = jnp.array([2, 3], dtype=jnp.float64)
    perturbations = jnp.array([pc, 0], dtype=jnp.float64)
    seed = jax.random.PRNGKey(0)
    new_features, new_rewards, new_perturbations = self.eagle._trim_pool(
        features_batch,
        rewards_batch,
        perturbations,
        best_reward=jnp.array(4.0),
        seed=seed,
    )

    np.testing.assert_array_almost_equal(
        new_features.continuous[0],
        features_batch.continuous[0],
        err_msg='Continuous features are not equal.',
    )
    np.testing.assert_array_almost_equal(
        new_features.categorical[0],
        features_batch.categorical[0],
        err_msg='Categorical features are not equal.',
    )
    self.assertTrue(
        np.all(
            np.not_equal(
                new_features.continuous[1], features_batch.continuous[1]
            )
        ),
        msg='Continuous features are not equal.',
    )
    self.assertTrue(
        np.all(
            np.not_equal(
                new_features.categorical[1], features_batch.categorical[1]
            )
        ),
        msg='Categorical features are not equal.',
    )

    np.testing.assert_array_equal(
        new_rewards,
        np.array([2, -np.inf], dtype=np.float64),
        err_msg='rewards are not equal.',
    )
    # The best firefly is never removed.
    np.testing.assert_array_almost_equal(
        new_perturbations,
        np.array([pc, pc], dtype=np.float64),
        err_msg='Perturbations are not equal.',
    )

  def test_create_strategy_from_factory(self):
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    root.add_float_param('x3', 0.0, 1.0)
    eagle_factory = eagle_strategy.VectorizedEagleStrategyFactory()
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    eagle = eagle_factory(converter)
    self.assertEqual(eagle.n_feature_dimensions.continuous, 3)
    self.assertEqual(eagle.n_feature_dimensions.categorical, 0)

  def test_optimize_with_eagle(self):

    eagle_factory = eagle_strategy.VectorizedEagleStrategyFactory()
    optimizer = vb.VectorizedOptimizerFactory(strategy_factory=eagle_factory)(
        self.converter
    )
    optimizer(
        score_fn=lambda x, _: -jnp.sum(x.continuous.padded_array, 1), count=1
    )

  def test_optimize_with_eagle_continuous_only(self):
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    root.add_float_param('x3', 0.0, 1.0)
    converter = converters.TrialToModelInputConverter.from_problem(problem)
    eagle_factory = eagle_strategy.VectorizedEagleStrategyFactory()
    optimizer = vb.VectorizedOptimizerFactory(strategy_factory=eagle_factory)(
        converter
    )
    n_parallel = 5
    results = optimizer(
        score_fn=lambda x, _: -jnp.sum(x.continuous.padded_array, axis=(1, 2)),
        count=1,
        n_parallel=n_parallel,
    )
    self.assertSequenceEqual(
        results.features.continuous.shape, (1, n_parallel, 3)
    )

  def test_optimize_with_eagle_padding(self):
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    root.add_float_param('x3', 0.0, 1.0)
    converter = converters.TrialToModelInputConverter.from_problem(
        problem,
        padding_schedule=padding.PaddingSchedule(
            num_trials=padding.PaddingType.POWERS_OF_2,
            num_features=padding.PaddingType.POWERS_OF_2,
        ),
    )
    eagle_factory = eagle_strategy.VectorizedEagleStrategyFactory()
    optimizer = vb.VectorizedOptimizerFactory(strategy_factory=eagle_factory)(
        converter
    )
    n_parallel = 2
    results = optimizer(
        score_fn=lambda x, _: -jnp.sum(x.continuous.padded_array, axis=(1, 2)),
        count=1,
        n_parallel=n_parallel,
    )
    self.assertSequenceEqual(
        results.features.continuous.shape, (1, n_parallel, 4)
    )

  @parameterized.parameters(
      dict(add_float_params=True, add_categorical_params=True),
      dict(add_float_params=True, add_categorical_params=False),
      dict(add_float_params=False, add_categorical_params=True),
  )
  def test_compute_feature_dimensions_from_converter(
      self, add_float_params: bool, add_categorical_params: bool
  ):
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    if add_float_params:
      root.add_float_param('x1', 0.0, 1.0)
      root.add_float_param('x2', 0.0, 1.0)
      root.add_float_param('x3', 0.0, 1.0)
    if add_categorical_params:
      root.add_categorical_param('c1', ['a', 'b', 'c'])
      root.add_categorical_param('c2', ['d', 'e', 'f', 'g'])
    converter = converters.TrialToModelInputConverter.from_problem(
        problem,
        padding_schedule=padding.PaddingSchedule(
            num_trials=padding.PaddingType.MULTIPLES_OF_10,
            num_features=padding.PaddingType.POWERS_OF_2,
        ),
    )
    feature_dimensions = (
        eagle_strategy.compute_feature_dimensions_from_converter(converter)
    )
    if add_float_params:
      self.assertEqual(feature_dimensions.n_feature_dimensions.continuous, 3)
      self.assertEqual(
          feature_dimensions.n_feature_dimensions_with_padding.continuous, 4
      )
    if add_categorical_params:
      self.assertEqual(feature_dimensions.n_feature_dimensions.categorical, 2)
      self.assertEqual(
          feature_dimensions.n_feature_dimensions_with_padding.categorical, 2
      )
      self.assertEqual(feature_dimensions.max_categorical_size, 4)
      self.assertEqual(feature_dimensions.categorical_sizes, [3, 4])

  @parameterized.parameters(
      {'num_prior_trials': 0},
      {'num_prior_trials': 3},
      {'num_prior_trials': 53},
  )
  def test_optimize_with_eagle_trials_padding(self, num_prior_trials):
    eagle_factory = eagle_strategy.VectorizedEagleStrategyFactory()
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    root.add_float_param('x3', 0.0, 1.0)
    root.add_categorical_param('c1', ['a', 'b', 'c'])
    root.add_categorical_param('c2', ['d', 'e', 'f', 'g'])

    rng = np.random.default_rng(seed=1)
    prior_features_continuous = rng.uniform(size=(num_prior_trials, 3))
    prior_features_categorical = rng.integers(
        low=0,
        high=[[3, 4]],
        size=(num_prior_trials, 2),
    )

    def score_fn(x, _):
      return jnp.sum(
          x.continuous.replace_fill_value(0).padded_array, axis=(1,)
      ) + jnp.sum(x.categorical.replace_fill_value(0).padded_array, axis=(1,))

    count = 3

    converter = converters.TrialToModelInputConverter.from_problem(
        problem,
    )
    optimizer = vb.VectorizedOptimizerFactory(strategy_factory=eagle_factory)(
        converter
    )
    prior_features = types.ModelInput(
        continuous=types.PaddedArray.as_padded(prior_features_continuous),
        categorical=types.PaddedArray.as_padded(prior_features_categorical),
    )
    results = optimizer(
        score_fn=score_fn,
        count=count,
        prior_features=None if num_prior_trials == 0 else prior_features,
    )

    padding_schedule = padding.PaddingSchedule(
        num_trials=padding.PaddingType.MULTIPLES_OF_10,
    )
    padding_converter = converters.TrialToModelInputConverter.from_problem(
        problem,
        padding_schedule=padding_schedule,
    )
    padding_optimizer = vb.VectorizedOptimizerFactory(
        strategy_factory=eagle_factory
    )(padding_converter)
    padded_prior_features = types.ModelInput(
        continuous=padding_schedule.pad_features(prior_features_continuous),
        categorical=padding_schedule.pad_features(prior_features_categorical),
    )
    padding_results = padding_optimizer(
        score_fn=score_fn,
        count=count,
        prior_features=None if num_prior_trials == 0 else padded_prior_features,
    )

    self.assertSequenceEqual(results.features.continuous.shape, (count, 1, 3))
    self.assertSequenceEqual(results.features.categorical.shape, (count, 1, 2))
    np.testing.assert_array_almost_equal(
        results.features.continuous,
        padding_results.features.continuous,
        decimal=5,
    )
    np.testing.assert_array_almost_equal(
        results.features.categorical,
        padding_results.features.categorical,
        decimal=5,
    )

  def test_factory(self):
    self.assertEqual(self.eagle.n_feature_dimensions.continuous, 2)
    self.assertEqual(self.eagle.n_feature_dimensions.categorical, 2)
    self.assertLen(self.eagle.categorical_sizes, 2)

  def test_sample_categorical_features(self):
    # features shouldn't have values in oov_mask, and have the structure of:
    # [c1,c1,c1,f1,c2,c2,c2,c2,d1]
    num_parallel = 3
    random_sampler = eagle_strategy.DefaultRandomSampler(self.converter)
    sampled_features = random_sampler(
        20,
        n_parallel=num_parallel,
        seed=jax.random.PRNGKey(0),
    )
    self.assertTrue(
        np.all(
            (sampled_features.continuous >= 0.0)
            & (sampled_features.continuous <= 1.0)
        )
    )
    self.assertTrue(
        np.all(
            sampled_features.categorical
            < np.array(self.eagle.categorical_sizes)
        )
    )

  def test_prior_trials(self):
    config = eagle_strategy.EagleStrategyConfig(
        visibility=1, gravity=1, pool_size=4, prior_trials_pool_pct=1.0
    )
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    root.add_categorical_param('c1', ['a', 'b', 'c'])
    converter = converters.TrialToModelInputConverter.from_problem(problem)

    prior_features_continuous = jnp.array(
        [[[1, -1]], [[2, 1]], [[3, 2]], [[4, 5]]], dtype=jnp.float32
    )
    prior_features = vb.VectorizedOptimizerInput(
        continuous=prior_features_continuous,
        categorical=jnp.array([[0], [2], [1], [1]])[:, jnp.newaxis],
    )
    prior_rewards = jnp.array([1, 2, 3, 4])
    eagle = eagle_strategy.VectorizedEagleStrategyFactory(
        eagle_config=config,
    )(converter=converter, suggestion_batch_size=2)
    init_state = eagle.init_state(
        jax.random.PRNGKey(0),
        prior_features=prior_features,
        prior_rewards=prior_rewards,
    )
    jax.tree_util.tree_map(
        lambda x, y: np.testing.assert_array_equal(y, jnp.flip(x, axis=0)),
        init_state.features,
        prior_features,
    )

  @parameterized.parameters(2, 10)
  def test_prior_trials_with_too_few_or_many_trials(self, n_prior_trials):
    # Test proper populating of the pool in case the number of prior trials is
    # greater or lesser than the pool size.
    config = eagle_strategy.EagleStrategyConfig(
        visibility=1, gravity=1, pool_size=4, prior_trials_pool_pct=1.0
    )
    problem = vz.ProblemStatement()
    root = problem.search_space.select_root()
    root.add_float_param('x1', 0.0, 1.0)
    root.add_float_param('x2', 0.0, 1.0)
    root.add_categorical_param('c1', ['a', 'b', 'c'])
    converter = converters.TrialToModelInputConverter.from_problem(problem)

    n_parallel = 3
    prior_features = vb.VectorizedOptimizerInput(
        continuous=jnp.asarray(np.random.randn(n_prior_trials, n_parallel, 2)),
        categorical=jnp.asarray(
            np.random.randint(3, size=(n_prior_trials, n_parallel, 1))
        ),
    )
    prior_rewards = jnp.asarray(np.random.randn(n_prior_trials))

    eagle = eagle_strategy.VectorizedEagleStrategyFactory(
        eagle_config=config,
    )(converter=converter, suggestion_batch_size=2)
    init_state = eagle.init_state(
        jax.random.PRNGKey(0),
        n_parallel=n_parallel,
        prior_features=prior_features,
        prior_rewards=prior_rewards,
    )
    self.assertEqual(
        init_state.features.continuous.shape, (config.pool_size, n_parallel, 2)
    )
    self.assertEqual(
        init_state.features.categorical.shape, (config.pool_size, n_parallel, 1)
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Custom Errors used in service."""


class AlreadyExistsError(ValueError):
  """The resource key already exists in the database."""

  pass


class NotFoundError(KeyError):
  """The resource key does not exist in the database."""

  pass


class ImmutableStudyError(ValueError):
  """The study is now immutable and cannot be modified."""

  pass


class ImmutableTrialError(ValueError):
  """The trial is now immutable and cannot be modified."""

  pass



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Gaussian Process models."""

import logging
from typing import Iterable, Optional, Union

import chex
import equinox as eqx
import jax
from jax import numpy as jnp
from tensorflow_probability.substrates import jax as tfp
from vizier._src.algorithms.designers.gp import acquisitions
from vizier._src.algorithms.designers.gp import transfer_learning as vtl
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier._src.jax.models import tuned_gp_models
from vizier.jax import optimizers
from vizier.utils import profiler

tfd = tfp.distributions


class GPTrainingSpec(eqx.Module):
  """Encapsulates all the information needed to train a singular GP model.

  Attributes:
    ard_optimizer: An `Optimizer` which should return a batch of hyperparameters
      to be ensembled.
    ard_rng: PRNG key for the ARD optimization.
    coroutine: The model coroutine.
    ensemble_size: If set, ensembles `ensemble_size` GP models together.
    ard_random_restarts: The number of random restarts.
  """

  ard_optimizer: optimizers.Optimizer[types.ParameterDict]
  ard_rng: jax.Array
  coroutine: sp.ModelCoroutine
  ensemble_size: int = eqx.field(static=True, default=1)
  ard_random_restarts: int = eqx.field(
      static=True, default=optimizers.DEFAULT_RANDOM_RESTARTS
  )


class GPState(eqx.Module):
  """A GP model and its training data. Implements `Predictive`.

  The data is kept around to deduce degrees of freedom and other related
  metrics. This implements `Predictive`, so that it and any of its dervied
  classes like `StackedResidualGP` can be used as a `Predictive` in
  `acquisitions.py`.
  """

  predictive: sp.UniformEnsemblePredictive
  data: types.ModelData

  def predict_with_aux(
      self, features: types.ModelInput
  ) -> tuple[tfd.Distribution, chex.ArrayTree]:
    return self.predictive.predict_with_aux(features)

  def num_hyperparameters(self) -> int:
    """Returns the number of hyperparameters optimized on `data`."""

    # For a GP model, this is feature dimensionality + 2
    # (length scales, amplitude, observation noise)
    # TODO: Compute this from the params returned by the ard
    # optimizer
    return (
        self.data.features.continuous.shape[1]  # (num_samples, num_features)
        + self.data.features.categorical.shape[1]  # (num_samples, num_features)
        + 2
    )


class StackedResidualGP(GPState):
  """GP that implements the `predictive` interface and contains stacked GPs.

  This GP handles sequential transfer learning. This holds one or no base
  (prior) GPs, along with a current top-level GP. The training process is such
  that the 'top' GP is trained on the residuals of the predictions from the
  base GP. The inference process is such that the predictions of the base
  GP and the 'top' GP are combined together.

  The base GP may also have its own base GPs and be a `StackedResidualGP`.
  """

  # `base_gp` refers to a GP trained and conditioned on previous data for
  # transfer learning. The top level GP is trained on the residuals from
  # `base_gp` on `data`.
  # If `None`, no transfer learning is used and all predictions happen through
  # `predictive`.
  base_gp: Optional[GPState] = None

  def predict_with_aux(
      self, features: types.ModelInput
  ) -> tuple[tfd.Distribution, chex.ArrayTree]:
    # Override the existing implementation of `predict_with_aux` to handle
    # combining `predictive` with `base_gp`.
    if not self.base_gp:
      return self.predictive.predict_with_aux(features)

    base_pred_dist, base_aux = self.base_gp.predict_with_aux(features)
    top_pred_dist, top_aux = self.predictive.predict_with_aux(features)

    base_pred = vtl.TransferPredictionState(
        pred=base_pred_dist,
        aux=base_aux,
        training_data_count=self.base_gp.data.labels.shape[0],
        num_hyperparameters=self.num_hyperparameters(),
    )
    top_pred = vtl.TransferPredictionState(
        pred=top_pred_dist,
        aux=top_aux,
        training_data_count=self.data.labels.shape[0],
        num_hyperparameters=self.num_hyperparameters(),
    )

    # TODO: Decide what to do with
    # `expected_base_stddev_mismatch` - currently set to default.
    comb_dist, aux = vtl.combine_predictions_with_aux(
        top_pred=top_pred, base_pred=base_pred
    )

    return comb_dist, aux


def get_vizier_gp_coroutine(
    data: types.ModelData,
    *,
    linear_coef: Optional[float] = None,
) -> sp.ModelCoroutine:
  """Gets a GP model coroutine.

  Args:
    data: The data used to the train the GP model
    linear_coef: If non-zero, uses a linear kernel with `linear_coef`
      hyperparameter.

  Returns:
    The model coroutine.
  """
  return tuned_gp_models.VizierGaussianProcess.build_model(
      data,
      linear_coef=linear_coef,
  ).coroutine


def _train_gp(spec: GPTrainingSpec, data: types.ModelData) -> GPState:
  """Trains a Gaussian Process model.

  1. Performs ARD to find the best model parameters.
  2. Pre-computes the Cholesky decomposition for the model.

  Args:
    spec: Spec required to train the GP. See `GPTrainingSpec` for more info.
    data: Data on which to train the GP.

  Returns:
    The trained GP model.
  """
  jax.monitoring.record_event(
      '/vizier/jax/designers/gp_bandit/train_gp', scope=profiler.current_scope()
  )

  jax.monitoring.record_event(
      '/vizier/jax/train_gp_with_data_shapes',
      **{
          'num_rows': data.features.categorical.shape[0],
          'num_categoricals': data.features.categorical.shape[1],
          'num_continuous': data.features.continuous.shape[1],
          'num_labels': (
              data.labels.shape[1] if data.labels.padded_array.ndim == 2 else 1
          ),
      },
  )
  model = sp.CoroutineWithData(spec.coroutine, data)

  # Optimize the parameters
  ard_rngs = jax.random.split(spec.ard_rng, spec.ard_random_restarts + 1)
  best_n = spec.ensemble_size or 1
  best_params, _ = spec.ard_optimizer(
      eqx.filter_jit(eqx.filter_vmap(model.setup))(ard_rngs[1:]),
      model.loss_with_aux,
      ard_rngs[0],
      constraints=model.constraints(),
      best_n=best_n,
  )
  if best_n == 1 and all(x.shape[0] == 1 for x in best_params.values()):
    best_params = jax.tree_util.tree_map(
        lambda x: jnp.squeeze(x, axis=0), best_params
    )
  best_models = sp.StochasticProcessWithCoroutine(
      coroutine=spec.coroutine, params=best_params
  )
  # Logging for debugging purposes.
  logging.info(
      'Best models: %s', eqx.tree_pformat(best_models, short_arrays=False)
  )
  predictive = sp.UniformEnsemblePredictive(
      eqx.filter_jit(best_models.precompute_predictive)(data)
  )
  return GPState(predictive=predictive, data=data)


@jax.jit
def _pred_mean(
    pred: acquisitions.Predictive, features: types.ModelInput
) -> types.Array:
  """Returns the mean of the predictions from `pred` on `features`.

  Workaround while `eqx.filter_jit(pred.pred_with_aux)(features)` is broken
  due to a bug in tensorflow probability.

  Args:
    pred: `Predictive` to predict with.
    features: Xs to predict on.

  Returns:
    Means of the predictions from `pred` on `features`.
  """
  return pred.predict_with_aux(features)[0].mean()


def train_stacked_residual_gp(
    base_gp: GPState,
    spec: GPTrainingSpec,
    data: types.ModelData,
) -> StackedResidualGP:
  """Trains a `StackedResidualGP`.

  Completes the following steps in order:
    1. Uses `base_gp` to predict on the `data`
    2. Computes the residuals from the above predictions
    3. Trains a top-level GP on the above residuals
    4. Returns a `StackedResidualGP` combining the base GP and newly-trained
    GP.

  Args:
    base_gp: The predictive to use as the base GP for the `StackedResidualGP`
      training.
    spec: Training spec for the top level GP.
    data: Training data for the top level GP.

  Returns:
    The trained `StackedResidualGP`.
  """
  # Compute the residuals of `data` as predicted by `base_gp`
  pred_means = _pred_mean(base_gp, data.features)

  has_no_padding = ~(
      data.features.continuous.is_missing[0]
      | data.features.categorical.is_missing[0]
      | data.labels.is_missing[0]
  )

  # Scope this to non-padded predictions only.
  pred_means_no_padding = pred_means[has_no_padding]
  residuals = (
      data.labels.unpad().reshape(pred_means_no_padding.shape)
      - pred_means_no_padding
  )

  # Train on the re-padded residuals
  residual_labels = types.PaddedArray.from_array(
      array=residuals,
      target_shape=data.labels.shape,
      fill_value=data.labels.fill_value,
  )
  data_with_residuals = types.ModelData(
      features=data.features, labels=residual_labels
  )

  top_gp = _train_gp(spec=spec, data=data_with_residuals)
  return StackedResidualGP(
      predictive=top_gp.predictive,
      data=top_gp.data,
      base_gp=base_gp,
  )


def train_gp(
    spec: Union[GPTrainingSpec, Iterable[GPTrainingSpec]],
    data: Union[types.ModelData, Iterable[types.ModelData]],
) -> GPState:
  """Trains a Gaussian Process model.

  If `spec` contains multiple elements, each will be used to train a
  `StackedResidualGP`, sequentially. The first entry will be used to train the
  first GP, and then subsequent GPs will be trained on the residuals from the
  previous GP. This process completes in the order that `spec` and `data are
  provided, such that `spec[0]` is the first GP trained and `spec[-1]` is the
  last GP trained.

  spec[-1] and data[-1] make up the top-level GP, and spec[:-1] and data[:-1]
  define the priors in context of transfer learning.

  Args:
    spec: Specification for how to train a GP model. If multiple specs are
      provided, transfer learning will train multiple models and combine into a
      single GP.
    data: Data on which to train GPs. NOTE: `spec` and `data` must be of the
      same shape. Trains a GP on `data[i]` with `spec[i]`.

  Returns:
    The trained GP model.
  """
  is_singleton_spec = isinstance(spec, GPTrainingSpec)
  is_singleton_data = isinstance(data, types.ModelData)
  if is_singleton_spec != is_singleton_data:
    raise ValueError(
        '`train_gp` expected the shapes of `spec` and `data`  to be identical.'
        f' Instead got `data` {data} but `spec` {spec}.'
    )

  if is_singleton_spec and is_singleton_data:
    return _train_gp(spec=spec, data=data)

  if len(spec) != len(data):
    raise ValueError(
        '`train_gp` expected the shapes of `spec` and `data` to be identical.'
        f' Instead got `spec` of length {len(spec)} but `data` of length'
        f' {len(data)}. `spec` was {spec} and `data` was {data}.'
    )

  curr_gp: Optional[GPState] = None
  for curr_spec, curr_data in zip(spec, data):
    if curr_gp is None:
      # We are on the first iteration.
      curr_gp = _train_gp(spec=curr_spec, data=curr_data)
    else:
      # Otherwise, we have a base GP to use - the GP trained on the last
      # iteration.
      curr_gp = train_stacked_residual_gp(
          base_gp=curr_gp,
          spec=curr_spec,
          data=curr_data,
      )

  if curr_gp is None:
    raise ValueError(
        f'Failed to train a GP with provided training spec: {spec} and'
        f' data: {data}. `curr_gp` was never updated. This should never happen.'
    )
  return curr_gp



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Cross-platform Vizier client interfaces.

Code written using these interfaces are compatible with OSS and Cloud Vertex
Vizier. Note importantly that subclasses may have more methods than what is
required by interfaces, and such methods are not cross compatible. Our
recommendation is to use type annotatations of `StudyInterface` or
`TrialInterface` wherever applicable.

Keywords:

#Materialize: The method returns a deep copy of the underlying pyvizier object.
Modifying the returned object does not update the Vizier service.
"""

import abc
from typing import Any, Iterator, List, Mapping, Optional, Type, TypeVar

from vizier import pyvizier as vz

_T = TypeVar('_T')


# TODO: Add more errors, and unit tests to ensure that
# all vizier clients have the same error handling logic.
class ResourceNotFoundError(LookupError):
  """Error raised by Vizier clients when resource is not found."""

  pass


class TrialInterface(abc.ABC):
  """Responsible for trial-level operations."""

  @property
  @abc.abstractmethod
  def id(self) -> int:
    """Identifier of the trial within the study.

    Ids are sorted in increasing order of creation time.
    """

  @property
  @abc.abstractmethod
  def parameters(self) -> Mapping[str, Any]:
    """#Materializes the parameters of the trial.

    The parameters are parsed to the external types. The values in the returned
    dict can be a list of values, if the search space is configured with
    indices.

    As a result, TrialInterface.parameters can be totally different
    from trial.materialize().parameters.
    """

  @abc.abstractmethod
  def delete(self) -> None:
    """Delete the Trial in Vizier service.

    There is currently no promise on how this object behaves after `delete()`.
    If you are sharing a Trial object in parallel processes, proceed with
    caution.
    """

  @abc.abstractmethod
  def update_metadata(self, delta: vz.Metadata) -> None:
    """Updates Trial metadata.

    New keys will be appended, while old keys will be
    updated with new values.

    Args:
      delta: Change in Metadata from the original.
    """

  @abc.abstractmethod
  def complete(
      self,
      measurement: Optional[vz.Measurement] = None,
      *,
      infeasible_reason: Optional[str] = None,
  ) -> Optional[vz.Measurement]:
    """Completes the trial and #materializes the measurement.

    * If `measurement` is provided, then Vizier writes it as the trial's final
    measurement and returns it.
    * If `infeasible_reason` is provided, `measurement` is not required but
    can still be specified.
    * If neither is provided, then Vizier selects an existing (intermediate)
    measurement to be the final measurement and returns it.

    Args:
      measurement: Final measurement.
      infeasible_reason: Infeasible reason for missing final measurement.

    Returns:
      The final measurement of the trial, or None if the trial is marked
      infeasible.

    Raises:
      ValueError: If neither `measurement` nor `infeasible_reason` is provided
        but the trial does not contain any intermediate measurements.
    """

  @abc.abstractmethod
  def check_early_stopping(self) -> bool:
    """Decide if the trial should stop.

    If the Trial is in ACTIVE state and an early stopping algorithm is
    configured fior the Study, the stopping algorithm makes a stopping decision.
    If the algorithm decides that the trial is not worth continuing, Vizier
    service moves it into `STOPPING` state.

    Then, this method returns True if the Trial is in `STOPPING` state.

    Returns:
      True if trial is already in STOPPING state or entered STOPPING as a
      result of this method invocation.
    """
    pass

  @abc.abstractmethod
  def stop(self) -> None:
    """Asks to change the trial status to STOPPING.

    Should be called only if the trial status is ACTIVE. If the trial is
    STOPPING or COMPLETED, this is a no-op.
    """
    pass

  @abc.abstractmethod
  def add_measurement(self, measurement: vz.Measurement) -> None:
    """Adds an intermediate measurement."""

  @abc.abstractmethod
  def materialize(self, *, include_all_measurements: bool = True) -> vz.Trial:
    """#Materializes the Trial.

    Args:
      include_all_measurements: If True, returned Trial includes all
        intermediate measurements. The `final_measurement` is always included if
        one exists.

    Returns:
      Trial object.
    """

  @property
  @abc.abstractmethod
  def study(self) -> 'StudyInterface':
    """Returns the Study that this Trial belongs to."""


class TrialIterable(abc.ABC):
  """Allows iterating through both `TrialInterface` and `vz.Trial`.

  TrialIterable satisfies Iterable[TrialInterface] Protocol which is not
  explicitly inherited (See https://peps.python.org/pep-0544/). In addition,
  it guarantees that `list(StudyInterface.trials().get())` is at least as fast
  as `[t.materialize() for t in StudyInterface.trials()]`.

  StudyInterface returns this object when trials are already materialized
  while processing a request. A typical implementation of TrialIterable.get()
  uses a generator of the materialized trials.
  """

  @abc.abstractmethod
  def __iter__(self) -> Iterator[TrialInterface]:
    """Returns an iterator of TrialInterfaces, which are clients."""

  @abc.abstractmethod
  def get(self) -> Iterator[vz.Trial]:
    """Returns Trial objects, which are local objects."""


class StudyInterface(abc.ABC):
  """Responsible for study-level operations."""

  @property
  @abc.abstractmethod
  def resource_name(self) -> str:
    """Globally unique identifier of the study in Vizier service."""

  @abc.abstractmethod
  def suggest(
      self, *, count: Optional[int] = None, client_id: str = 'default_client_id'
  ) -> List[TrialInterface]:
    """Returns Trials to be evaluated by client_id.

    Args:
      count: Number of suggestions.
      client_id: When new Trials are generated, their `client_id` field is
        populated with this client_id. The Vizier service first looks for
        existing ACTIVE Trials that are assigned to `client_id`, before
        generating new ones.

    Returns:
      Trials.
    """
    # TODO: Add `parameters` argument. This will allow
    # `suggest` to be used with context.

  # TODO: Request does not play well with boolean or discrete
  # integer parameters.
  @abc.abstractmethod
  def request(self, suggestion: vz.TrialSuggestion) -> TrialInterface:
    """Request a trial to be suggested in the future.

    Requested trials are "queued" in the vizier database. Next time a client
    requests for a new suggestion, it receives requested trials before any
    algorithm-generated suggestions.

    Args:
      suggestion: Suggestion to be requested.
    """

  @abc.abstractmethod
  def delete(self) -> None:
    """Delete the Study in Vizier service.

    There is currently no promise on how this object behaves after `delete()`.
    If you are sharing a Study object in parallel processes, proceed with
    caution.
    """

  @abc.abstractmethod
  def update_metadata(self, delta: vz.Metadata) -> None:
    """Updates StudyConfig metadata.

    New keys will be appended, while old keys will be
    updated with new values.

    Args:
      delta: Change in Metadata from the original.
    """

  @abc.abstractmethod
  def add_trial(self, trial: vz.Trial) -> TrialInterface:
    """Adds a trial to the Study. Allows warm-starting.

    Args:
      trial: Trial to be added.

    Returns:
      Trial client.

    Raises:
      ValueError: If the trial is not within the search space.
    """

  @abc.abstractmethod
  def trials(
      self, trial_filter: Optional[vz.TrialFilter] = None
  ) -> TrialIterable:
    """Fetches a collection of trials. Default uses vz.TrialFilter()."""

  @abc.abstractmethod
  def get_trial(self, uid: int) -> TrialInterface:
    """Fetches a single trial.

    Args:
      uid: Unique identifier of the trial within study.

    Returns:
      Trial.

    Raises:
      ResourceNotFoundError: If trial does not exist.
    """

  @abc.abstractmethod
  def optimal_trials(self, *, count: Optional[int] = None) -> TrialIterable:
    """Returns (pareto) optimal trial(s). Can be multiple Trial(s).

    Args:
      count: If provided, returns that many best Trials, breaking ties by
        earlier Trial id.
    """

  @abc.abstractmethod
  def materialize_problem_statement(self) -> vz.ProblemStatement:
    """#Materializes the problem statement."""

  @classmethod
  @abc.abstractmethod
  # TODO: Make the function have position only parameter
  # once cloud vizier support python 3.8+ version.
  def from_resource_name(cls: Type[_T], name: str) -> _T:
    """Fetches an existing study from the Vizier service.

    Args:
      name: Globally unique identifier of the study.

    Returns:
      Study.

    Raises:
      ResourceNotFoundError: If study does not exist.
    """

  @abc.abstractmethod
  def set_state(self, state: vz.StudyState) -> None:
    """Sets the state of the study.

    Args:
      state: New state of the study.
    """

  # TODO: Make this method purely abstract.
  def materialize_state(self) -> vz.StudyState:
    """#Materializes the study state."""
    raise NotImplementedError(
        f'materialize_state is not implemented in {type(self)}!'
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Convenience classes for configuring Vizier Study Configs and Search Spaces.

This module contains several classes, used to access/build Vizier StudyConfig
protos:
  * `StudyConfig` class is the main class, which:
  1) Allows to easily build Vizier StudyConfig protos via a convenient
     Python API.
  2) Can be initialized from an existing StudyConfig proto, to enable easy
     Pythonic accessors to information contained in StudyConfig protos,
     and easy field editing capabilities.

  * `SearchSpace` and `SearchSpaceSelector` classes deals with Vizier search
    spaces. Both flat spaces and conditional parameters are supported.
"""

import collections
import copy
import enum
from typing import Dict, List, Optional, Sequence, Tuple, Union

import attr
from vizier._src.pyvizier.oss import automated_stopping
from vizier._src.pyvizier.oss import metadata_util
from vizier._src.pyvizier.oss import proto_converters
from vizier._src.pyvizier.shared import base_study_config
from vizier._src.pyvizier.shared import common
from vizier._src.pyvizier.shared import parameter_config
from vizier._src.pyvizier.shared import trial
from vizier._src.service import constants
from vizier._src.service import study_pb2


################### PyTypes ###################
# Possible types for trial parameter values after cast to external types.
# TODO: Define this in _src/shared/
ParameterValueSequence = Union[
    trial.ParameterValueTypes,
    Sequence[int],
    Sequence[float],
    Sequence[str],
    Sequence[bool],
]

################### Enums ###################


class Algorithm(enum.Enum):
  """Valid Values for StudyConfig.Algorithm."""
  # Let Vizier choose algorithm. Currently defaults to GP_UCB_PE.
  ALGORITHM_UNSPECIFIED = 'ALGORITHM_UNSPECIFIED'
  # Gaussian Process UCB with Pure Exploration.
  GP_UCB_PE = 'GP_UCB_PE'
  # Gaussian Process Bandit.
  GAUSSIAN_PROCESS_BANDIT = 'GAUSSIAN_PROCESS_BANDIT'
  # Grid search within the feasible space.
  GRID_SEARCH = 'GRID_SEARCH'
  # Grid search, but with parameters and values shuffled.
  SHUFFLED_GRID_SEARCH = 'SHUFFLED_GRID_SEARCH'
  # Random search within the feasible space.
  RANDOM_SEARCH = 'RANDOM_SEARCH'
  # Quasi-random search using Halton sequences.
  QUASI_RANDOM_SEARCH = 'QUASI_RANDOM_SEARCH'
  # NSGA2 (https://ieeexplore.ieee.org/document/996017).
  NSGA2 = 'NSGA2'
  # BOCS (https://arxiv.org/abs/1806.08838) only applicable to boolean search
  # spaces.
  BOCS = 'BOCS'
  # Harmonica (https://arxiv.org/abs/1706.00764) only applicable to boolean
  # search spaces.
  HARMONICA = 'HARMONICA'
  # CMA-ES (https://arxiv.org/abs/1604.00772) for DOUBLE search spaces only
  CMA_ES = 'CMA_ES'
  # Eagle Strategy (https://doi.org/10.1007/978-3-642-04944-6_14).
  EAGLE_STRATEGY = 'EAGLE_STRATEGY'


class ObservationNoise(enum.Enum):
  """Valid Values for StudyConfig.ObservationNoise."""

  OBSERVATION_NOISE_UNSPECIFIED = (
      study_pb2.StudySpec.ObservationNoise.OBSERVATION_NOISE_UNSPECIFIED
  )
  LOW = study_pb2.StudySpec.ObservationNoise.LOW
  HIGH = study_pb2.StudySpec.ObservationNoise.HIGH


################### Main Class ###################
#
# A StudyConfig object can be initialized:
# (1) From a StudyConfig proto using StudyConfig.from_proto():
#     study_config_proto = study_pb2.StudySpec(...)
#     study_config = pyvizier.StudyConfig.from_proto(study_config_proto)
#     # Attributes can be modified.
#     study_config.metadata['metadata_key'] = 'metadata_value'
#     new_proto = study_config.to_proto()
#
# (2) By directly calling __init__ and setting attributes:
#     study_config = pyvizier.StudyConfig(
#       metric_information=[pyvizier.MetricInformation(
#         name='accuracy', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE)],
#       search_space=vz.SearchSpace.from_proto(proto),
#     )
#     # OR:
#     study_config = pyvizier.StudyConfig()
#     study_config.metric_information.append(
#        pyvizier.MetricInformation(
#          name='accuracy', goal=pyvizier.ObjectiveMetricGoal.MAXIMIZE))
#
#     # Since building a search space is more involved, get a reference to the
#     # search space, and add parameters to it.
#     root = study_config.search_space.root
#     root.add_float_param('learning_rate', 0.001, 1.0,
#       scale_type=pyvizier.ScaleType.LOG)
#


@attr.define(frozen=False, init=True, slots=True, kw_only=True)
class StudyConfig(base_study_config.ProblemStatement):
  """A builder and wrapper for study_pb2.StudySpec proto."""

  algorithm: str = attr.field(
      init=True,
      validator=attr.validators.instance_of((Algorithm, str)),
      converter=lambda x: x.value if isinstance(x, enum.Enum) else x,
      on_setattr=[attr.setters.convert, attr.setters.validate],
      default='ALGORITHM_UNSPECIFIED',
      kw_only=True)

  pythia_endpoint: Optional[str] = attr.field(
      init=True,
      validator=attr.validators.optional(attr.validators.instance_of(str)),
      on_setattr=[attr.setters.convert, attr.setters.validate],
      default=None,
      kw_only=True)

  observation_noise: ObservationNoise = attr.field(
      init=True,
      validator=attr.validators.instance_of(ObservationNoise),
      on_setattr=attr.setters.validate,
      default=ObservationNoise.OBSERVATION_NOISE_UNSPECIFIED,
      kw_only=True)

  automated_stopping_config: Optional[
      automated_stopping.AutomatedStoppingConfig] = attr.field(
          init=True,
          default=None,
          validator=attr.validators.optional(
              attr.validators.instance_of(
                  automated_stopping.AutomatedStoppingConfig)),
          on_setattr=attr.setters.validate,
          kw_only=True)

  # An internal representation as a StudyConfig proto.
  # If this object was created from a StudyConfig proto, a copy of the original
  # proto is kept, to make sure that unknown proto fields are preserved in
  # round trip serialization.
  # TODO: Fix the broken proto validation.
  _study_config: study_pb2.StudySpec = attr.field(
      init=True,
      factory=study_pb2.StudySpec,
      kw_only=True)

  # Public attributes, methods and properties.
  @classmethod
  def pythia_endpoint_metadata(cls, pythia_endpoint: str) -> common.Metadata:
    """Returns the MetaData for updating the pythia endpoint."""
    metadata = common.Metadata()
    metadata.ns(constants.PYTHIA_ENDPOINT_NAMESPACE)[
        constants.PYTHIA_ENDPOINT_KEY
    ] = pythia_endpoint
    return metadata

  @classmethod
  def from_proto(cls, proto: study_pb2.StudySpec) -> 'StudyConfig':
    """Converts a StudyConfig proto to a StudyConfig object.

    Args:
      proto: StudyConfig proto.

    Returns:
      A StudyConfig object.
    """
    algorithm = proto.algorithm

    metric_information = base_study_config.MetricsConfig(
        sorted(
            [
                proto_converters.MetricInformationConverter.from_proto(m)
                for m in proto.metrics
            ],
            key=lambda x: x.name,
        )
    )

    oneof_name = proto.WhichOneof('automated_stopping_spec')
    if not oneof_name:
      automated_stopping_config = None
    else:
      automated_stopping_config = (
          automated_stopping.AutomatedStoppingConfig.from_proto(
              getattr(proto, oneof_name)
          )
      )

    metadata = common.Metadata()
    for kv in proto.metadata:
      metadata.abs_ns(common.Namespace.decode(kv.ns))[kv.key] = (
          kv.proto if kv.HasField('proto') else kv.value
      )

    # Store the pythia_endpoint as a property for convenience.
    pythia_endpoint = None
    try:
      pythia_endpoint = metadata.ns(constants.PYTHIA_ENDPOINT_NAMESPACE)[
          constants.PYTHIA_ENDPOINT_KEY
      ]
    except KeyError:
      pass  # Pythia endpoint doesn't exist.

    return cls(
        search_space=proto_converters.SearchSpaceConverter.from_proto(proto),
        algorithm=algorithm,
        pythia_endpoint=pythia_endpoint,
        metric_information=metric_information,
        observation_noise=ObservationNoise(proto.observation_noise),
        automated_stopping_config=automated_stopping_config,
        study_config=copy.deepcopy(proto),
        metadata=metadata)

  def to_proto(self) -> study_pb2.StudySpec:
    """Serializes this object to a StudyConfig proto."""
    proto = copy.deepcopy(self._study_config)
    proto.algorithm = self.algorithm
    proto.observation_noise = self.observation_noise.value

    del proto.metrics[:]
    proto.metrics.extend(
        proto_converters.MetricsConfigConverter.to_protos(
            self.metric_information))

    del proto.parameters[:]
    proto.parameters.extend(
        proto_converters.SearchSpaceConverter.parameter_protos(
            self.search_space))

    if self.automated_stopping_config is not None:
      auto_stop_proto = self.automated_stopping_config.to_proto()
      if isinstance(auto_stop_proto,
                    study_pb2.StudySpec.DefaultEarlyStoppingSpec):
        proto.default_stopping_spec.CopyFrom(auto_stop_proto)

    # The internally stored proto already contains metadata.
    proto.ClearField('metadata')
    for ns in self.metadata.namespaces():
      ns_string = ns.encode()
      ns_layer = self.metadata.abs_ns(ns)
      for key, value in ns_layer.items():
        metadata_util.assign(proto, key=key, ns=ns_string, value=value)
    if self.pythia_endpoint is not None:
      ns = common.Namespace([constants.PYTHIA_ENDPOINT_NAMESPACE])
      metadata_util.assign(
          proto,
          key=constants.PYTHIA_ENDPOINT_KEY,
          ns=ns.encode(),
          value=self.pythia_endpoint,
          mode='insert_or_assign',
      )
    return proto

  def _trial_to_external_values(
      self, pytrial: trial.Trial
  ) -> Dict[str, Union[float, int, str, bool]]:
    """Returns the trial paremeter values cast to external types."""
    parameter_values: Dict[str, Union[float, int, str]] = {}
    external_values: Dict[str, Union[float, int, str, bool]] = {}
    # parameter_configs is a list of Tuple[parent_name, ParameterConfig].
    parameter_configs: List[
        Tuple[Optional[str], parameter_config.ParameterConfig]
    ] = [(None, p) for p in self.search_space.parameters]
    remaining_parameters = copy.deepcopy(pytrial.parameters)
    # Traverse the conditional tree using a BFS.
    while parameter_configs and remaining_parameters:
      parent_name, pc = parameter_configs.pop(0)
      parameter_configs.extend(
          (pc.name, child) for child in pc.child_parameter_configs
      )
      if pc.name not in remaining_parameters:
        continue
      if parent_name is not None:
        # This is a child parameter. If the parent was not seen,
        # skip this parameter config.
        if parent_name not in parameter_values:
          continue
        parent_value = parameter_values[parent_name]
        if parent_value not in pc.matching_parent_values:
          continue
      parameter_values[pc.name] = remaining_parameters[pc.name].value
      if pc.external_type is None:
        external_value = remaining_parameters[pc.name].value
      else:
        external_value = remaining_parameters[pc.name].cast(pc.external_type)  # pytype: disable=wrong-arg-types
      external_values[pc.name] = external_value
      remaining_parameters.pop(pc.name)
    return external_values

  def trial_parameters(
      self, proto: study_pb2.Trial) -> Dict[str, ParameterValueSequence]:
    """Returns the trial values, cast to external types, if they exist.

    Args:
      proto:

    Returns:
      Parameter values dict: cast to each parameter's external_type, if exists.
      NOTE that the values in the dict may be a Sequence as opposed to a single
      element.

    Raises:
      ValueError: If the trial parameters do not exist in this search space.
      ValueError: If the trial contains duplicate parameters.
    """
    pytrial = proto_converters.TrialConverter.from_proto(proto)
    return self._pytrial_parameters(pytrial)

  def _pytrial_parameters(
      self, pytrial: trial.Trial
  ) -> Dict[str, ParameterValueSequence]:
    """Returns the trial values, cast to external types, if they exist.

    Args:
      pytrial:

    Returns:
      Parameter values dict: cast to each parameter's external_type, if exists.
      NOTE that the values in the dict may be a Sequence as opposed to a single
      element.

    Raises:
      ValueError: If the trial parameters do not exist in this search space.
      ValueError: If the trial contains duplicate parameters.
    """
    trial_external_values: Dict[str, Union[float, int, str, bool]] = (
        self._trial_to_external_values(pytrial))
    if len(trial_external_values) != len(pytrial.parameters):
      raise ValueError('Invalid trial for this search space: failed to convert '
                       'all trial parameters: {}'.format(pytrial))

    # Combine multi-dimensional parameter values to a list of values.
    trial_final_values: Dict[str, ParameterValueSequence] = {}
    # multi_dim_params: Dict[str, List[Tuple[int, ParameterValueSequence]]]
    multi_dim_params = collections.defaultdict(list)
    for name in trial_external_values:
      base_index = parameter_config.SearchSpaceSelector.parse_multi_dimensional_parameter_name(
          name
      )
      if base_index is None:
        trial_final_values[name] = trial_external_values[name]
      else:
        base_name, index = base_index
        multi_dim_params[base_name].append((index, trial_external_values[name]))
    for name in multi_dim_params:
      multi_dim_params[name].sort(key=lambda x: x[0])
      trial_final_values[name] = [x[1] for x in multi_dim_params[name]]

    return trial_final_values

  def trial_metrics(self,
                    proto: study_pb2.Trial,
                    *,
                    include_all_metrics=False) -> Dict[str, float]:
    """Returns the trial's final measurement metric values.

    If the trial is not completed, or infeasible, no metrics are returned.
    By default, only metrics configured in the StudyConfig are returned
    (e.g. only objective and safety metrics).

    Args:
      proto:
      include_all_metrics: If True, all metrics in the final measurements are
        returned. If False, only metrics configured in the StudyConfig are
        returned.

    Returns:
      Dict[metric name, metric value]
    """
    pytrial = proto_converters.TrialConverter.from_proto(proto)
    return self._pytrial_metrics(
        pytrial, include_all_metrics=include_all_metrics)

  def _pytrial_metrics(
      self, pytrial: trial.Trial, *, include_all_metrics=False
  ) -> Dict[str, float]:
    """Returns the trial's final measurement metric values.

    If the trial is not completed, or infeasible, no metrics are returned.
    By default, only metrics configured in the StudyConfig are returned
    (e.g. only objective and safety metrics).

    Args:
      pytrial:
      include_all_metrics: If True, all metrics in the final measurements are
        returned. If False, only metrics configured in the StudyConfig are
        returned.

    Returns:
      Dict[metric name, metric value]
    """
    configured_metrics = [m.name for m in self.metric_information]

    metrics: Dict[str, float] = {}
    if pytrial.is_completed and not pytrial.infeasible:
      if pytrial.final_measurement is None:
        return metrics
      for name in pytrial.final_measurement.metrics:
        if (include_all_metrics or
            (not include_all_metrics and name in configured_metrics)):
          # Special case: Measurement always adds an empty metric by default.
          # If there is a named single objective in study_config, drop the empty
          # metric.
          if not name and self.single_objective_metric_name != name:
            continue
          metrics[name] = pytrial.final_measurement.metrics[name].value
    return metrics



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from vizier import pyvizier as vz
from vizier._src.algorithms.designers import scalarization
from vizier._src.algorithms.designers import scalarizing_designer
from vizier._src.algorithms.designers import unsafe_as_infeasible_designer
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy
from vizier._src.algorithms.testing import test_runners
from vizier.testing import test_studies
from absl.testing import absltest


class UnsafeAsInfeasibleDesignerTest(absltest.TestCase):

  def test_unsafe_eagle(self):
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.extend([
        vz.MetricInformation(
            name='metric1', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        ),
        vz.MetricInformation(
            name='metric_safe',
            goal=vz.ObjectiveMetricGoal.MAXIMIZE,
            safety_threshold=0.2,
        ),
    ])

    def eagle_designer_factory(ps, seed):
      return eagle_strategy.EagleStrategyDesigner(
          problem_statement=ps, seed=seed
      )

    safe_eagle = unsafe_as_infeasible_designer.UnsafeAsInfeasibleDesigner(
        problem, eagle_designer_factory
    )

    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=6,
            batch_size=5,
            verbose=1,
            validate_parameters=True,
        ).run_designer(safe_eagle),
        30,
    )

  def test_unsafe_multi_objective_eagle(self):
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.extend([
        vz.MetricInformation(
            name='metric1', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        ),
        vz.MetricInformation(
            name='metric2', goal=vz.ObjectiveMetricGoal.MINIMIZE
        ),
        vz.MetricInformation(
            name='metric_safe',
            goal=vz.ObjectiveMetricGoal.MAXIMIZE,
            safety_threshold=0.2,
        ),
    ])

    def eagle_designer_factory(ps, seed):
      return eagle_strategy.EagleStrategyDesigner(
          problem_statement=ps, seed=seed
      )

    def scalarized_eagle_factory(ps, seed):
      return scalarizing_designer.create_gaussian_scalarizing_designer(
          ps,
          eagle_designer_factory,
          scalarization.HyperVolumeScalarization,
          num_ensemble=3,
          seed=seed,
      )

    safe_eagle = unsafe_as_infeasible_designer.UnsafeAsInfeasibleDesigner(
        problem, scalarized_eagle_factory
    )

    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=6,
            batch_size=5,
            verbose=1,
            validate_parameters=True,
        ).run_designer(safe_eagle),
        30,
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Includes both tools for defining a flax model and a library of models."""
# pylint: disable=unused-import

from vizier._src.jax.models.tuned_gp_models import VizierGaussianProcess
from vizier._src.jax.stochastic_process_model import Constraint
from vizier._src.jax.stochastic_process_model import get_constraints
from vizier._src.jax.stochastic_process_model import InitFn
from vizier._src.jax.stochastic_process_model import ModelCoroutine
from vizier._src.jax.stochastic_process_model import ModelParameter
from vizier._src.jax.stochastic_process_model import ModelParameterGenerator
from vizier._src.jax.stochastic_process_model import StochasticProcessModel
from vizier._src.jax.stochastic_process_model import VectorToArrayTree



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Defines core components for tuner integration."""

import collections
import contextlib
import datetime
import typing
from typing import Any, Optional, Sequence

from absl import logging
import attr
import pyglove as pg
from vizier import pyvizier as vz
from vizier._src.pyglove import constants
from vizier._src.pyglove import converters
from vizier.client import client_abc


def _trial_status_legacy_value(status: vz.TrialStatus) -> str:
  # PENDING was renamed to ACTIVE.
  if status == vz.TrialStatus.ACTIVE:
    return 'PENDING'
  return status.value


class VizierTrial(pg.tuning.Trial):
  """Override Trial to lazy load DNA and metadata upon access.

  When we construct a `Trial` object, it doesn't pop up DNA, measurements and
  metadata from vizier trial proto immediately. This is because that a study
  may consists of thousands of trials, if we load them at construction time, it
  would take minutes, which is not acceptable. So we made the `Trial` object
  lazily load these properties upon access, reducing the construction time into
  a few seconds.
  """

  # Here we explicitly override the __init__ method managed by PyGlove,
  # for we want to pass in `converter` and `trial` which are not managed by
  # PyGlove.
  @pg.explicit_method_override
  def __init__(
      self, converter: converters.VizierConverter, trial: vz.Trial, **kwargs
  ):
    completed_time = (
        int(trial.completion_time.timestamp()) if trial.completion_time else 0
    )
    super().__init__(
        dna=pg.DNA(None),
        id=trial.id,
        description=trial.description,
        final_measurement=converter.to_tuner_measurement(
            trial.final_measurement
        ),
        status=_trial_status_legacy_value(trial.status),
        created_time=int(trial.creation_time.timestamp()),
        completed_time=completed_time,
        infeasible=trial.infeasible,
        **kwargs,
    )
    self._converter = converter
    self._trial = trial

  @property
  def dna(self) -> pg.DNA:
    """Returns lazy loaded DNA."""
    temp_dna = self.sym_init_args.dna
    if temp_dna.value is None and not temp_dna.children:
      self.sym_init_args.dna = self._converter.to_dna(self._trial)
    return self.sym_init_args.dna

  @property
  def metadata(self) -> dict[str, Any]:
    """Returns lazy loaded metadata."""
    if not self.sym_init_args.metadata and self._trial:
      self.sym_init_args.metadata = converters.get_pyglove_metadata(self._trial)
    return self.sym_init_args.metadata

  @property
  def related_links(self) -> dict[str, str]:
    """Returns lazy loaded related links."""
    if not self.sym_init_args.related_links and self._trial:
      self.sym_init_args.related_links = dict(
          self._trial.metadata.ns(constants.METADATA_NAMESPACE).ns(
              constants.RELATED_LINKS_SUBNAMESPACE
          )
      )
    return self.sym_init_args.related_links

  @property
  def measurements(self) -> list[pg.tuning.Measurement]:
    """Returns lazy loaded measurements."""
    if not self.sym_init_args.measurements:
      self.sym_init_args.measurements = [
          self._converter.to_tuner_measurement(m)
          for m in self._trial.measurements
      ]
    return self.sym_init_args.measurements

  def format(self, *args, **kwargs):
    """Fetch lazy bound properties before print."""
    # NOTE(daiyip): `format` depends on the symbolic attributes to generate
    # the string representation. Since the following symbolic attributes are
    # lazily assigned upon property accesses, we prefetch them before calling
    # the `format`. Otherwise, the symbolic attributes are just default values
    # set at __init__ time.
    _, _, _, _ = self.dna, self.measurements, self.metadata, self.related_links
    return super().format(*args, **kwargs)


def _parse_namespace_from_key(
    encoded_key: str, default_ns: vz.Namespace
) -> tuple[vz.Namespace, str]:
  """From ':ns:key' to (ns, key)."""
  ns_and_key = tuple(vz.Namespace.decode(encoded_key))
  if not ns_and_key:
    raise ValueError(
        f'String did not parse into namespace and key: {encoded_key}'
    )
  elif len(ns_and_key) == 1:
    return (default_ns, ns_and_key[-1])
  else:
    return (vz.Namespace(ns_and_key[:-1]), ns_and_key[-1])


class Feedback(pg.tuning.Feedback):
  """Tuning feedback for a vizier trial."""

  def __init__(
      self,
      vizier_trial: client_abc.TrialInterface,
      converter: converters.VizierConverter,
  ):
    """Constructor.

    Args:
      vizier_trial: Vizier trial (cross-platform).
      converter: Vizier-Pyglove converter.
    """
    super().__init__(converter.metrics_to_optimize)
    self._converter = converter
    self._trial_client = vizier_trial
    self._trial = self._trial_client.materialize()
    self._dna_spec = converter.dna_spec
    self._discard_reward = 'reward' not in converter.metrics_to_optimize

  @property
  def id(self) -> int:
    """Gets Trial ID as ID."""
    return self._trial_client.id

  @property
  def dna(self) -> pg.DNA:
    """Gets DNA of current trial."""
    return self._converter.to_dna(self._trial)

  def get_trial(self) -> pg.tuning.Trial:
    """Gets current trial with all fields up-to-date."""
    self._trial = self._trial_client.materialize()
    return VizierTrial(self._converter, self._trial)

  @property
  def checkpoint_to_warm_start_from(self) -> Optional[str]:
    """Gets checkpoint path to warm start from. Refreshes `_trial`."""
    return None

  @contextlib.contextmanager
  def _maybe_race_condition(self, message: str):
    """Raise race condition error when error message matches with regex."""
    try:
      yield
    # TODO: once pyvizier expose common error types, we should
    # change `Exception` to a narrower error type.
    except Exception as e:  # pylint:disable=broad-except
      if message in str(e):
        raise pg.tuning.RaceConditionError(str(e)) from e
      else:
        raise

  def _add_measurement(
      self,
      reward: Optional[float],
      metrics: dict[str, float],
      step: int,
      checkpoint_path: Optional[str],
      elapse_secs: float,
  ) -> None:
    """Reports tuning measurement to the pg.tuning."""
    if reward is not None and not self._discard_reward:
      metrics |= {'reward': reward}
    with self._maybe_race_condition('Measurements can only be added to'):
      self._trial_client.add_measurement(
          vz.Measurement(metrics, elapsed_secs=elapse_secs, steps=step)
      )

  def set_metadata(self, key: str, value: Any, per_trial: bool = True) -> None:
    """Sets metadata for current trial or current sampling."""
    md = vz.Metadata()
    md.ns(constants.METADATA_NAMESPACE)[key] = pg.to_json_str(value)
    if per_trial:
      self._trial_client.update_metadata(md)
    else:
      self._trial_client.study.update_metadata(md)

  def get_metadata(self, key: str, per_trial: bool = True) -> Optional[Any]:
    """Gets metadata for current trial or current sampling.

    Args:
      key: A key to the Trial or StudyConfig metadata.  Vizier treats this as
        {namespace}:{key} where colons in {key} are escaped with a backslash,
        and {namespace} is encoded per vizier.pyvizier.Namespace.encode (i.e.
        colons are escaped and namespace components are separated by colons).
        But for the special case of the empty namespace, this simplifies to be
        just the key string.
      per_trial: True if you want to see per-trial metadata for the current
        Trial; false for study-related metadata.

    Returns:
      A metadata item, interpreted as JSON format.
    """
    abs_ns, key_in_ns = _parse_namespace_from_key(
        key, default_ns=vz.Namespace([constants.METADATA_NAMESPACE])
    )
    if per_trial:
      value = self._trial.metadata.abs_ns(abs_ns).get(key_in_ns, None)
    else:
      value = (
          self._trial_client.study.materialize_problem_statement()
          .metadata.abs_ns(abs_ns)
          .get(key_in_ns, None)
      )
    return pg.from_json_str(value) if value is not None else None

  def add_link(self, name: str, url: str) -> None:
    """Adds related link."""
    md = vz.Metadata()
    md.ns(constants.METADATA_NAMESPACE).ns(
        constants.RELATED_LINKS_SUBNAMESPACE
    )[name] = url
    self._trial_client.update_metadata(md)

  def done(
      self,
      metadata: Optional[dict[str, Any]] = None,
      related_links: Optional[dict[str, str]] = None,
  ) -> None:
    """Marks current tuning trial as done, and export final object."""
    metadata = metadata or {}
    related_links = related_links or {}
    for key, value in metadata.items():
      self.set_metadata(key, value)
    for key, value in related_links.items():
      self.add_link(key, value)
    self._trial_client.complete()
    self._trial = self._trial_client.materialize()

  def skip(self, reason: Optional[str] = None) -> None:
    """Skips current trial without providing feedback to the controller."""
    self._trial_client.complete(infeasible_reason=reason or 'skipped')

  def should_stop_early(self) -> bool:
    """Tells whether this trial should be stopped."""
    return self._trial_client.check_early_stopping()

  def end_loop(self) -> None:
    """Ends current search loop."""
    self._trial_client.study.set_state(vz.StudyState.ABORTED)


@attr.define(repr=False)
class Result(pg.tuning.Result):
  """Vizier tuner progress."""

  _converter: converters.VizierConverter = attr.field()
  _problem: vz.ProblemStatement = attr.field()
  _study: client_abc.StudyInterface = attr.field()
  _metadata: pg.Dict = attr.field()
  _best_trial: pg.tuning.Trial = attr.field()
  _trials: Sequence[pg.tuning.Trial] = attr.field()
  _num_trials_by_status: dict[str, int] = attr.field()
  _last_update_time: datetime.datetime = attr.field(
      factory=datetime.datetime.now
  )

  @classmethod
  def from_study(cls, study: client_abc.StudyInterface):
    # Recover DNA spec
    logging.info('from study..')
    problem = study.materialize_problem_statement()
    logging.info('got metadata.')
    metadata = converters.get_pyglove_study_metadata(problem)
    converter = converters.VizierConverter.from_problem(problem)

    logging.info('Getting trials..')
    best_trial = VizierTrial(converter, next(study.optimal_trials().get()))
    tuner_trials = [VizierTrial(converter, t) for t in study.trials().get()]
    logging.info('Got trials...')
    num_trials_by_status = dict(
        collections.Counter(t.status for t in tuner_trials)
    )

    return cls(
        converter,
        problem,
        study,
        metadata,
        best_trial,
        tuner_trials,
        num_trials_by_status,
    )

  @property
  def last_updated(self) -> datetime.datetime:
    """Last update time."""
    return self._last_update_time

  @property
  def is_active(self) -> bool:
    """Returns whether tuner is active."""
    state = self._study.materialize_state()
    active = state == vz.StudyState.ACTIVE
    logging.info('is_active was called. state:%s, active:%s', state, active)
    return active

  @property
  def metadata(self) -> dict[str, Any]:
    """Gets metadata for current sampling."""
    return self._metadata

  @property
  def best_trial(self) -> Optional[pg.tuning.Trial]:
    """Returns the best trial."""
    return self._best_trial

  @property
  def trials(self) -> Sequence[pg.tuning.Trial]:
    """Returns trials."""
    return self._trials

  def format(
      self,
      compact: bool = False,
      verbose: bool = True,
      root_indent: int = 0,
      **kwargs,
  ):
    # Return summary.
    status_field = pg.tuning.Trial.__schema__.get_field('status')
    assert status_field is not None
    possible_status = set(
        typing.cast(pg.typing.Enum, status_field.value).values
    )
    study = f'{self._study.resource_name}'
    # TODO: Add link to the study.
    status = {
        s: f'{self._num_trials_by_status[s]}/{len(self._trials)}'
        for s in possible_status
        if s in self._num_trials_by_status
    }
    json_repr = dict(study=study, status=status)
    if self._best_trial:
      json_repr['best_trial'] = dict(
          id=self._best_trial.id,
          reward=self._best_trial.final_measurement.reward,
          step=self._best_trial.final_measurement.step,
          dna=self._best_trial.dna.format(compact=True),
      )
    return pg.format(json_repr, compact, False, root_indent, **kwargs)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Eagle Strategy Optimizer.

Implements a variation of Eagle Strategy without the Levy random walk, aka
Firefly Algorithm (FA) for the purpose of optimizing a given objective function.

Reference: Yang XS. (2009) Firefly Algorithms for Multimodal Optimization.
In: Stochastic Algorithms: Foundations and Applications (SAGA) 2009.
DOI: https://doi.org/10.1007/978-3-642-04944-6_14

Firefly Algorithm Summary
=========================
FA is a genetic algorithm that maintains a pool of fireflies. Each
firefly emits a light whose intensity is non-decreasing in (or simply equal
to) the objective value. Each iteration, a firefly chases after a brighter
firefly, but the brightness it perceives decreases in distance. This allows
multiple "clusters" to form, as opposed to all fireflies collapsing to a
single point. Not included in the original algorithm, we added "repulsion" which
in addition to the "attraction" forces, meaning fireflies move towards the
bright spots as well as away from the dark spots. We also support non-decimal
parameter types (categorical, discrete, integer), and treat them uniquely when
computing distance, adding pertrubation, and mutating fireflies.

For more details, see the linked paper.

OSS Vizier Implementation Summary
=================================
The fireflies are stored in three JAX arrays: features, rewards, perturbations.
Each iteration we mutate 'batch_size' fireflies to generate new features. The
new features are evaluated on the objective function to obtain their associated
rewards and update the pool where improvement was obtained, and decrease the
perturbation factor otherwise.

If the firefly's perturbation reaches the `perturbation_lower_bound` threshold
it's removed and replaced with new random features.

For performance consideration, the 'pool size' is a multiplier of the
'batch size', and so each iteration the pool is sliced to obtain the current
fireflies to be mutated.

Passing prior trials is supported for knowledge-transfering from previous runs
of the optimizer. Prior trials are used to populate the pool with fireflies
which are closer to the optimum.

Note that the strategy assumes that the search space is not conditional.

Example
=======
# Construct the optimizer with eagle strategy.
optimizer = VectorizedOptimizerFactory(
    VectorizedEagleStrategyFactory(),
)
# Run the optimization.
trials = optimizer.optimize(problem_statement, objective_function)
"""
import dataclasses
# pylint: disable=g-long-lambda

import enum
import logging
import math
from typing import Callable, Optional, Tuple

import attr
from flax import struct
import jax
from jax import numpy as jnp
from tensorflow_probability.substrates import jax as tfp
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.jax import types
from vizier.pyvizier import converters

tfd = tfp.distributions


@enum.unique
class MutateNormalizationType(enum.IntEnum):
  """The force normalization mode. Use IntEnum for JIT compatibility."""

  MEAN = 0
  RANDOM = 1
  UNNORMALIZED = 2


@struct.dataclass
class EagleStrategyConfig:
  """Eagle Strategy optimizer config.

  Attributes:
    visibility: The sensetivity to distance between flies when computing pulls.
    gravity: The maximum amount of attraction pull.
    negative_gravity: The maximum amount of repulsion pull.
    perturbation: The default amount of noise for perturbation.
    categorical_perturbation_factor: A factor to apply on categorical params.
    pure_categorical_perturbation_factor: A factor on purely categorical space.
    prob_same_category_without_perturbation: Baseline probability of selecting
      the same category.
    perturbation_lower_bound: The threshold below flies are removed from pool.
    penalize_factor: The perturbation decrease for unsuccessful flies.
    pool_size_exponent: An exponent for computing pool size based on search
      space size.
    pool_size: An optional way to set the pool size. If not 0, this takes
      precedent over the programatic pool size computation.
    max_pool_size: Ceiling on pool size.
    mutate_normalization_type: The type of force mutation normalizatoin used for
      damping down the force intensity to stay within reasonable bounds.
    normalization_scale: A mutation force scale-factor to control intensity.
    prior_trials_pool_pct: The percentage of the pool populated by prior trials.
  """

  # Visibility
  visibility: float = 0.45
  # Gravity
  gravity: float = 1.5
  negative_gravity: float = 0.008
  # Perturbation
  perturbation: float = 0.16
  categorical_perturbation_factor: float = 1.0
  pure_categorical_perturbation_factor: float = 30
  prob_same_category_without_perturbation: float = 0.98
  # Penalty
  perturbation_lower_bound: float = 7e-5
  penalize_factor: float = 7e-1
  # Pool size
  pool_size_exponent: float = struct.field(pytree_node=False, default=1.2)
  pool_size: int = struct.field(pytree_node=False, default=0)
  max_pool_size: int = struct.field(pytree_node=False, default=100)
  # Force normalization mode
  mutate_normalization_type: MutateNormalizationType = struct.field(
      pytree_node=False,
      default=MutateNormalizationType.MEAN,
  )
  # Multiplier factor when using normalized modes
  normalization_scale: float = 0.5
  # The percentage of the firefly pool to be populated with prior trials
  prior_trials_pool_pct: float = struct.field(pytree_node=False, default=0.96)


@dataclasses.dataclass(frozen=True)
class FeatureDimensions:
  """Dimensions of the features used by Eagle."""

  # The number of categories for each categorical feature.
  categorical_sizes: list[int]
  # The number of feature dimensions with feature padding.
  n_feature_dimensions_with_padding: types.ContinuousAndCategorical[int]
  # The number of feature dimensions without feature padding.
  n_feature_dimensions: types.ContinuousAndCategorical[int]

  @property
  def max_categorical_size(self) -> int:
    """The maximum number of categories across the categorical features."""
    return max(self.categorical_sizes) if self.categorical_sizes else 0


def compute_feature_dimensions_from_converter(
    converter: converters.TrialToModelInputConverter,
) -> FeatureDimensions:
  """Computes feature dimensions for Eagle from a converter.

  Args:
    converter: A converter from trials to the input to the acquisition function.

  Returns:
    A FeatureDimensions object that contains the feature dimensions information
    for Eagle.
  """
  empty_features = converter.to_features([])
  n_feature_dimensions_with_padding = types.ContinuousAndCategorical[int](
      continuous=empty_features.continuous.shape[-1],
      categorical=empty_features.categorical.shape[-1],
  )
  n_feature_dimensions = types.ContinuousAndCategorical(
      continuous=len(converter.output_specs.continuous),
      categorical=len(converter.output_specs.categorical),
  )

  categorical_sizes = []
  for spec in converter.output_specs.categorical:
    categorical_sizes.append(spec.bounds[1])
  extra_categories = (
      n_feature_dimensions_with_padding.categorical
      - n_feature_dimensions.categorical
  )
  # Pads the categorical sizes with 0s to match the number of categorical
  # features with padding.
  categorical_sizes = categorical_sizes + [0] * extra_categories
  return FeatureDimensions(
      categorical_sizes=categorical_sizes,
      n_feature_dimensions_with_padding=n_feature_dimensions_with_padding,
      n_feature_dimensions=n_feature_dimensions,
  )


class DefaultProjection(vb.Projection):
  """Default projection for vectorized optimizer."""

  def __init__(
      self,
      converter: converters.TrialToModelInputConverter,
  ):
    del converter

  def __call__(
      self,
      x: vb.VectorizedOptimizerInput,
  ) -> vb.VectorizedOptimizerInput:
    # TODO: The range of features is not always [0, 1].
    #   Specifically, for features that are single-point, it can be [0, 0]; we
    #   also want this code to be aware of the feature's bounds to enable
    #   contextual bandit operation.  Note that if a parameter's bound changes,
    #   we might also want to change the firefly noise or normalizations.
    return vb.VectorizedOptimizerInput(
        continuous=jnp.clip(x.continuous, 0.0, 1.0), categorical=x.categorical
    )


class DefaultRandomSampler(vb.RandomSampler):
  """Default random sampler for vectorized optimizer."""

  def __init__(
      self,
      converter: converters.TrialToModelInputConverter,
  ):
    feature_dimensions = compute_feature_dimensions_from_converter(converter)
    self._max_categorical_size = feature_dimensions.max_categorical_size
    self._categorical_sizes = feature_dimensions.categorical_sizes
    self._continuous_padded_dim = (
        feature_dimensions.n_feature_dimensions_with_padding.continuous
    )

  def __call__(
      self,
      num_samples: int,
      n_parallel: int,
      seed: jax.Array,
  ) -> vb.VectorizedOptimizerInput:
    cont_seed, cat_seed = jax.random.split(seed)

    if self._max_categorical_size > 0:
      # Appends a new axis to `self._categorical_sizes` to make it broadcastable
      # with the shape of jnp.arange(self._max_categorical_size) in the
      # comparison below.
      sizes = jnp.array(self._categorical_sizes)[:, jnp.newaxis]
      logits = jnp.where(
          jnp.arange(self._max_categorical_size) < sizes,
          0.0,
          -jnp.inf,
      )  # shape (sizes.shape[0], max_categorical_size)
      random_categorical_features = (
          tfd.Categorical(logits=logits)
          .sample((num_samples, n_parallel), seed=cat_seed)
          .astype(types.INT_DTYPE)
      )
    else:
      random_categorical_features = jnp.zeros(
          [num_samples, n_parallel, 0], types.INT_DTYPE
      )
    return types.ContinuousAndCategoricalArray(
        continuous=jax.random.uniform(
            cont_seed,
            shape=(
                num_samples,
                n_parallel,
                self._continuous_padded_dim,
            ),
        ),
        categorical=random_categorical_features,
    )


@attr.define(frozen=True)
class VectorizedEagleStrategyFactory(vb.VectorizedStrategyFactory):
  """Eagle strategy factory."""

  eagle_config: EagleStrategyConfig = attr.field(factory=EagleStrategyConfig)
  random_sampler_factory: Callable[
      [converters.TrialToModelInputConverter], vb.RandomSampler
  ] = attr.field(default=DefaultRandomSampler)
  projection_factory: Callable[
      [converters.TrialToModelInputConverter], vb.Projection
  ] = attr.field(default=DefaultProjection)

  def __call__(
      self,
      converter: converters.TrialToModelInputConverter,
      suggestion_batch_size: Optional[int] = None,
  ) -> "VectorizedEagleStrategy":
    """Create a new vectorized eagle strategy.

    In order to create the strategy a converter has to be passed, which the
    strategy will then store a pointer to. The strategy uses the converter to
    get information about the original Vizier Parameters and their relation to
    the array indices (which index belong to which Vizier Parameter). This is
    useful for example for sampling CATEGORICAL features.

    Arguments:
      converter: TrialToArrayConverter that matches the converter used in
        computing the objective / acuisition function.
      suggestion_batch_size: The batch_size of the returned suggestion array.

    Returns:
      A new instance of VectorizedEagleStrategy.
    """
    valid_types = [
        converters.NumpyArraySpecType.DISCRETE,
        converters.NumpyArraySpecType.CONTINUOUS,
    ]
    if any(
        spec.type not in valid_types
        for spec in (
            list(converter.output_specs.continuous)
            + list(converter.output_specs.categorical)
        )
    ):
      raise ValueError("Only DISCRETE/CONTINUOUS parameters are supported!")

    feature_dimensions = compute_feature_dimensions_from_converter(converter)
    n_features = (
        feature_dimensions.n_feature_dimensions.continuous
        + feature_dimensions.n_feature_dimensions.categorical
    )
    pool_size = self.eagle_config.pool_size
    if pool_size == 0:
      pool_size = 10 + int(
          0.5 * n_features + n_features**self.eagle_config.pool_size_exponent
      )
      pool_size = min(pool_size, self.eagle_config.max_pool_size)
      if suggestion_batch_size is not None:
        # If the batch_size was set, ensure pool_size is multiply of batch_size.
        pool_size = int(
            math.ceil(pool_size / suggestion_batch_size) * suggestion_batch_size
        )
    logging.info("Pool size: %d", pool_size)
    if suggestion_batch_size is None:
      # This configuration updates all the fireflies in each iteration.
      suggestion_batch_size = pool_size
    # Use priors to populate Eagle state
    # pytype: disable=wrong-arg-types  # jnp-type
    return VectorizedEagleStrategy(
        n_feature_dimensions=feature_dimensions.n_feature_dimensions,
        n_feature_dimensions_with_padding=(
            feature_dimensions.n_feature_dimensions_with_padding
        ),
        random_sampler=self.random_sampler_factory(converter),
        projection=self.projection_factory(converter),
        batch_size=suggestion_batch_size,
        config=self.eagle_config,
        pool_size=pool_size,
        categorical_sizes=jnp.array(feature_dimensions.categorical_sizes),
        max_categorical_size=feature_dimensions.max_categorical_size,
        dtype=converter._impl.dtype,
    )
    # pytype: enable=wrong-arg-types


@struct.dataclass
class VectorizedEagleStrategyState:
  """Container for Eagle strategy state."""

  iterations: jax.Array  # Scalar integer.
  features: vb.VectorizedOptimizerInput  # (pool_size, n_parallel, n_features).
  rewards: jax.Array  # Shape (pool_size,).
  best_reward: jax.Array  # Scalar float.
  perturbations: jax.Array  # Shape (pool_size,).


def _compute_features_dist_squared(
    x_batch: vb.VectorizedOptimizerInput, x_pool: vb.VectorizedOptimizerInput
) -> jax.Array:
  """Computes squared distance between features (or parallel feature batches).

  The distance computed is the squared Euclidean distance between each feature
  in the batch and each feature in the pool.

  To avoid materializing an array of size (batch_size, pool_size, n_features),
  the distances are computed efficiently using the formula (a-b)**2 = a**2 +
  b**2 - 2ab and broadcasting.

  Reference:
  https://github.com/tensorflow/probability/blob/r0.24/tensorflow_probability/python/math/psd_kernels/internal/util.py#L191

  Args:
    x_batch: A batch of features to be compared to the pool.
    x_pool: The pool of features to be compared to the batch.

  Returns:
    Array of size (batch_size, pool_size) with the squared distances.
  """
  dist = jnp.zeros([], dtype=x_batch.continuous.dtype)
  if x_batch.continuous.size > 0:
    x_batch_cont = jnp.reshape(
        x_batch.continuous, (x_batch.continuous.shape[0], -1)
    )
    x_pool_cont = jnp.reshape(
        x_pool.continuous, (x_pool.continuous.shape[0], -1)
    )
    continuous_dists = (
        jnp.sum(x_batch_cont**2, axis=-1, keepdims=True)
        + jnp.sum(x_pool_cont**2, axis=-1)
        - 2.0 * jnp.matmul(x_batch_cont, x_pool_cont.T)
    )  # shape (batch_size, pool_size)
    dist = dist + continuous_dists

  if x_batch.categorical.size > 0:
    x_batch_cat = jnp.reshape(
        x_batch.categorical, (x_batch.categorical.shape[0], -1)
    )
    x_pool_cat = jnp.reshape(
        x_pool.categorical, (x_pool.categorical.shape[0], -1)
    )
    categorical_diffs = (x_batch_cat[..., jnp.newaxis, :] != x_pool_cat).astype(
        x_batch.continuous.dtype
    )
    dist = dist + jnp.sum(categorical_diffs, axis=-1)
  return dist


def _mask_flip(
    prior_features: vb.VectorizedOptimizerInput, prior_rewards: types.Array
) -> Tuple[vb.VectorizedOptimizerInput, types.Array]:
  """Flips the ordering of the elements in `prior_rewards` and `prior_features`.

  Args:
    prior_features: Prior features to be flipped.
    prior_rewards: Prior rewards to be flipped.

  Returns:
    A tuple of flipped prior features and prior rewards such that all elements
    corresponding to -inf entries in `prior_rewards` are at the end, while all
    other elements have the opposite order. For example, if `prior_rewards` is
    [1, -jnp.inf, 3, -jnp.inf ,2],  `flipped_prior_rewards` will be
    [2, 3, 1, -jnp.inf, -jnp.inf].
  """
  mask = jnp.invert(jnp.isneginf(prior_rewards))
  indices = jnp.flip(
      jnp.argsort(jnp.where(mask, jnp.arange(prior_rewards.shape[0]), -1))
  )
  flipped_prior_features = jax.tree_util.tree_map(
      lambda x: x[indices], prior_features
  )
  flipped_prior_rewards = prior_rewards[indices]
  return flipped_prior_features, flipped_prior_rewards


@struct.dataclass
class VectorizedEagleStrategy(
    vb.VectorizedStrategy[VectorizedEagleStrategyState]
):
  """Eagle strategy implementation for maximization problem based on Numpy.

  Attributes:
    config: The Eagle strategy configuration.
    n_features: The number of features.
    batch_size: The number of suggestions generated at each suggestion call.
    pool_size: The total number of flies in the pool.
  """

  n_feature_dimensions: types.ContinuousAndCategorical[int]
  categorical_sizes: jax.Array
  n_feature_dimensions_with_padding: types.ContinuousAndCategorical[int] = (
      struct.field(pytree_node=False)
  )
  max_categorical_size: int = struct.field(pytree_node=False)
  pool_size: int = struct.field(pytree_node=False)
  dtype: jnp.dtype = struct.field(pytree_node=False)
  random_sampler: vb.RandomSampler = struct.field(pytree_node=False)
  projection: vb.Projection = struct.field(pytree_node=False)
  batch_size: Optional[int] = struct.field(pytree_node=False, default=None)
  config: EagleStrategyConfig = struct.field(
      default_factory=EagleStrategyConfig
  )

  def init_state(
      self,
      seed: jax.Array,
      n_parallel: int = 1,
      *,
      prior_features: Optional[vb.VectorizedOptimizerInput] = None,
      prior_rewards: Optional[types.Array] = None,
  ) -> VectorizedEagleStrategyState:
    """Initializes the state."""
    if prior_features is not None and prior_rewards is not None:
      if prior_features.continuous.shape[1] != n_parallel:
        raise ValueError(
            "`prior_features.continuous` dimension 1 "
            f"({prior_features.continuous.shape[1]}) "
            f"doesn't match n_parallel ({n_parallel})!"
        )
      if prior_features.categorical.shape[1] != n_parallel:
        raise ValueError(
            "`prior_features.categorical` dimension 1 "
            f"({prior_features.categorical.shape[1]}) "
            f"doesn't match n_parallel ({n_parallel})!"
        )
      init_features = self._populate_pool_with_prior_trials(
          seed, prior_features, prior_rewards
      )
    else:
      init_features = self.random_sampler(
          self.pool_size,
          n_parallel=n_parallel,
          seed=seed,
      )
    # pytype: disable=wrong-arg-types  # jnp-type
    return VectorizedEagleStrategyState(
        iterations=jnp.array(0),
        features=init_features,
        rewards=jnp.ones(self.pool_size) * -jnp.inf,
        best_reward=-jnp.inf,
        perturbations=jnp.ones(self.pool_size) * self.config.perturbation,
    )
    # pytype: enable=wrong-arg-types

  def _populate_pool_with_prior_trials(
      self,
      seed: jax.Array,
      prior_features: types.ContinuousAndCategoricalArray,
      prior_rewards: types.Array,
  ) -> types.ContinuousAndCategoricalArray:
    """Populate the pool with prior trials.

    A portion of the pool is first populated with random features based on
    'prior_trials_pool_pct', then the rest of the flies are populated by
    sequentially iterate over the prior trials, finding the cloest firefly in
    the pool and replace it if the reward is better.

    Args:
      seed: Random seed.
      prior_features: (n_prior_features, n_parallel, features_count)
      prior_rewards: (n_prior_features, )

    Returns:
      initial_features
    """
    if prior_features is None or prior_rewards is None:
      raise ValueError("One of prior features / prior rewards wasn't provided!")

    if prior_features.continuous is not None:
      continuous_obs, _, continuous_dim = prior_features.continuous.shape
      if continuous_obs != prior_rewards.shape[0]:
        raise ValueError(
            f"prior continuous features shape ({continuous_obs}) doesn't match"
            f" prior rewards shape ({prior_rewards.shape[0]})!"
        )
      expected_dim = self.n_feature_dimensions_with_padding.continuous
      if continuous_dim != expected_dim:
        raise ValueError(
            f"prior continuous features shape ({continuous_dim}) doesn't match "
            f"n_features {expected_dim}!"
        )
    if prior_features.categorical is not None:
      categorical_obs, _, categorical_dim = prior_features.categorical.shape
      if categorical_obs != prior_rewards.shape[0]:
        raise ValueError(
            f"prior categorical features shape ({categorical_obs}) doesn't "
            f"match prior rewards shape ({prior_rewards.shape[0]})!"
        )
      expected_dim = self.n_feature_dimensions_with_padding.categorical
      if categorical_dim != expected_dim:
        raise ValueError(
            f"prior categorical features shape ({categorical_dim}) doesn't"
            f" match n_features {expected_dim}!"
        )
    if len(prior_rewards.shape) > 1:
      raise ValueError("prior rewards is expected to be 1D array!")

    # Reverse the order of prior trials to assign more weight to recent trials.
    flipped_prior_features, flipped_prior_rewards = _mask_flip(
        prior_features, prior_rewards
    )

    n_parallel = flipped_prior_features.continuous.shape[1]

    pool_random_features = self.random_sampler(
        self.pool_size,
        n_parallel,
        seed,
    )
    n_random_flies = int(
        self.pool_size * (1 - self.config.prior_trials_pool_pct)
    )
    # The pool is configured to have at least this many random trials.
    init_features = jax.tree_util.tree_map(
        lambda x: x[:n_random_flies, :, :], pool_random_features
    )
    pool_left_space = self.pool_size - n_random_flies
    # When the number of prior trials is smaller than configured, we fill in
    # random trials.
    random_features = jax.tree_util.tree_map(
        lambda x: x[n_random_flies:, :, :], pool_random_features
    )

    # Starts with the most recent `pool_left_space` prior trials as the chosen
    # set of trials, and loops through the remaining prior trials. At each
    # iteration, if the remaining trial has a better reward than its closest
    # neighbor in the chosen set, it replaces the closest neighbor. Note that
    # `pool_left_space` can be larger than the number of prior trials, in which
    # case the chosen set contains all prior trials.
    features = jax.tree_util.tree_map(
        lambda x: x[:pool_left_space], flipped_prior_features
    )
    rewards = flipped_prior_rewards[:pool_left_space]

    def _loop_body(i, args):
      features, rewards = args
      ind = jnp.argmin(
          _compute_features_dist_squared(
              jax.tree_util.tree_map(
                  lambda x: x[i][jnp.newaxis],
                  flipped_prior_features,
              ),
              features,
          ),
          axis=-1,
      )[0]
      return jax.lax.cond(
          rewards[ind] < flipped_prior_rewards[i],
          lambda: (
              jax.tree_util.tree_map(
                  lambda f, pf: f.at[ind].set(pf[i]),
                  features,
                  flipped_prior_features,
              ),
              rewards.at[ind].set(flipped_prior_rewards[i]),
          ),
          lambda: (features, rewards),
      )

    # TODO: Use a vectorized method to populate the pool and avoid
    # the for-loop.
    features, _ = jax.lax.fori_loop(
        lower=pool_left_space,
        upper=prior_rewards.shape[0],
        body_fun=_loop_body,
        init_val=(features, rewards),
    )

    num_chosen_trials = rewards.shape[0]
    # Replaces padded trials in the chosen trials with random trials.
    features = jax.tree_util.tree_map(
        lambda x, y: jnp.where(
            jnp.isneginf(rewards)[:, jnp.newaxis, jnp.newaxis],
            x[:num_chosen_trials, :, :],
            y,
        ),
        random_features,
        features,
    )
    # Then ensures the chosen set of trials has exactly `pool_left_space` trials
    # by filling in random trials.
    features = jax.tree_util.tree_map(
        lambda x, y: jnp.concatenate([x, y[num_chosen_trials:, :, :]]),
        features,
        random_features,
    )

    return jax.tree_util.tree_map(
        lambda x, y: jnp.concatenate([x, y]), init_features, features
    )

  @property
  def suggestion_batch_size(self) -> int | None:
    """The number of suggestions returned at each call of 'suggest'."""
    return self.batch_size

  def suggest(
      self,
      seed: jax.Array,
      state: VectorizedEagleStrategyState,
      n_parallel: int = 1,
  ) -> vb.VectorizedOptimizerInput:
    """Suggest new mutated and perturbed features.

    After initializing, at each call a `batch_size` of firefiles is taken
    from the pool and mutated to generate new features. The batch of fireflies
    is chosen based on the current iteration number. The mutation is done
    by applying pulling/pushing forces from all the fireflies in the pool.

    The update rule is:
    x' <-- x' + gravity * sum_i exp(-visibility * ||x'-x_i||^2) (x_i-x') + noise

    Args:
      seed: Random seed.
      state: Current strategy state.
      n_parallel: Number of points that the acquisition function maps to a
        single value. This arg may be greater than 1 if a parallel acquisition
        function (qEI, qUCB) is used; otherwise it should be 1.

    Returns:
      suggested batch features: (batch_size, n_parallel, n_features)
    """
    # Take the batch of fireflies based on which new features are generated.
    batch_id = state.iterations % (self.pool_size // self.batch_size)
    start = batch_id * self.batch_size
    features_batch = jax.tree_util.tree_map(
        lambda f: jax.lax.dynamic_slice_in_dim(f, start, self.batch_size),
        state.features,
    )
    rewards_batch = jax.lax.dynamic_slice_in_dim(
        state.rewards, start, self.batch_size
    )
    perturbations_batch = jax.lax.dynamic_slice_in_dim(
        state.perturbations, start, self.batch_size
    )
    features_seed, perturbations_seed = jax.random.split(seed)

    def _mutate_features(features_batch_):
      perturbations = self._create_random_perturbations(
          perturbations_batch, n_parallel, perturbations_seed
      )
      return self._create_features(
          state.features,
          state.rewards,
          features_batch_,
          rewards_batch,
          perturbations,
          features_seed,
      )

    # If the strategy is still initializing, return the random/prior features.
    new_features = jax.lax.cond(
        state.iterations < self.pool_size // self.batch_size,
        lambda x: x,
        _mutate_features,
        features_batch,
    )

    return self.projection(new_features)

  def _create_features(
      self,
      features: vb.VectorizedOptimizerInput,
      rewards: jax.Array,
      features_batch: vb.VectorizedOptimizerInput,
      rewards_batch: jax.Array,
      perturbations_batch: types.ContinuousAndCategoricalArray,
      seed: jax.Array,
  ) -> vb.VectorizedOptimizerInput:
    """Create new batch of mutated and perturbed features.

    The combined forces (pull/push) induced by the firefly pool is normalized to
    ensure it's contained and doesn't throw the firefly too far. Mathematically,
    the normalization guarantees that the combined normalized force is within
    the simplex spanned by the unnormalized forces and therefore within bounds.

    Args:
      features: (pool_size, n_parallel, n_features)
      rewards: (pool_size,)
      features_batch: (batch_size, n_parallel, n_features)
      rewards_batch: (batch_size,)
      perturbations_batch: (batch_size,)
      seed: Random seed.

    Returns:
      batch features: (batch_size, n_parallel, n_features)
    """
    # Compute the pairwise squared distances between the batch and the
    # pool features.
    dists = _compute_features_dist_squared(
        features_batch, features
    )  # shape: (batch_size, pool_size)

    # Compute the scaled direction for applying pull between two flies.
    # scaled_directions[i,j] := direction of force applied by fly 'j' on fly
    # 'i'. Note that to compute 'directions' we might perform subtract with
    # removed flies having value of -np.inf. We might even subtract between two
    # removed flies which will result in np.nan. Both cases are handled when
    # computing the actual feature changes by applying a relevant mask.
    directions = rewards - rewards_batch[:, jnp.newaxis]
    scaled_directions = jnp.where(
        directions >= 0.0, self.config.gravity, -self.config.negative_gravity
    )  # shape: (batch_size, pool_size)

    # Get the number of non-padded features.
    n_feature_dimensions = sum(
        jax.tree_util.tree_leaves(self.n_feature_dimensions)
    )
    # Normalize the distance by the number of feature dimensions.
    force = jnp.exp(
        -self.config.visibility * dists / n_feature_dimensions * 10.0
    )
    # Compute the directed force each pool firefly applies on batch fireflies.
    scaled_force = scaled_directions * force  # shape: (batch_size, pool_size)
    # Handle removed fireflies without updated rewards.
    finite_ind = jnp.isfinite(rewards).astype(scaled_force.dtype)
    # Ignore fireflies that were removed from the pool.
    scaled_force = scaled_force * finite_ind
    # Separate forces to pull and push so to normalize them separately.
    scaled_pulls = jnp.maximum(scaled_force, 0.0)
    scaled_push = jnp.minimum(scaled_force, 0.0)

    seed, categorical_seed = jax.random.split(seed)
    if self.config.mutate_normalization_type == MutateNormalizationType.MEAN:
      # Divide the push / pull forces by the number of participating fireflies.
      # Also multiply by normalization_scale.
      # pytype: disable=wrong-arg-types  # jnp-type
      norm_scaled_pulls = self.config.normalization_scale * jnp.nan_to_num(
          scaled_pulls / jnp.sum(scaled_pulls > 0.0, axis=1, keepdims=True), 0
      )
      norm_scaled_push = self.config.normalization_scale * jnp.nan_to_num(
          scaled_push / jnp.sum(scaled_push < 0.0, axis=1, keepdims=True), 0
      )
      # pytype: enable=wrong-arg-types
    elif self.config.mutate_normalization_type == (
        MutateNormalizationType.RANDOM
    ):
      # Create random matrices and normalize each row, s.t. the sum is 1.
      pull_seed, push_seed = jax.random.split(seed)
      scaled_pulls_pos = scaled_pulls > 0
      pull_rand_matrix = (
          jax.random.uniform(pull_seed, shape=scaled_pulls.shape)
          * scaled_pulls_pos
      )
      pull_weight_matrix = pull_rand_matrix / jnp.sum(
          pull_rand_matrix, axis=1, keepdims=True
      )
      push_rand_matrix = (
          jax.random.uniform(push_seed, shape=scaled_pulls.shape)
          * scaled_pulls_pos
      )
      push_weight_matrix = push_rand_matrix / jnp.sum(
          push_rand_matrix, axis=1, keepdims=True
      )
      # Normalize pulls/pulls by the weight matrices and multiply by
      # normalization_scale.
      norm_scaled_pulls = (
          self.config.normalization_scale * scaled_pulls * pull_weight_matrix
      )
      norm_scaled_push = (
          self.config.normalization_scale * scaled_push * push_weight_matrix
      )
    elif self.config.mutate_normalization_type == (
        MutateNormalizationType.UNNORMALIZED
    ):
      # Doesn't normalize the forces. Use this option with caution.
      norm_scaled_pulls = scaled_pulls
      norm_scaled_push = scaled_push
    else:
      raise ValueError(
          "Unsupported mutate normalization type:"
          f" {self.config.mutate_normalization_type}"
      )

    # Sums the normalized forces (pull/push) of all fireflies.
    scale = norm_scaled_pulls + norm_scaled_push
    flat_features = jnp.reshape(
        features.continuous, (features.continuous.shape[0], -1)
    )  # shape: (pool_size, n_features)
    flat_features_batch = jnp.reshape(
        features_batch.continuous, (features_batch.continuous.shape[0], -1)
    )  # shape: (batch_size, n_features)

    # Compute the feature changes without materializing a large matrix:
    # batch_features_i += sum_j (scale_{ij} * (features_j - batch_features_i))
    # shape: (batch_size, pool_size) @ (pool_size, n_features) - (batch_size,
    # n_features) * (batch_size, 1).
    features_changes_continuous = jnp.matmul(
        scale, flat_features
    ) - flat_features_batch * jnp.sum(scale, axis=-1, keepdims=True)

    new_features_continuous = (
        features_batch.continuous
        + jnp.reshape(
            features_changes_continuous, features_batch.continuous.shape
        )
        + perturbations_batch.continuous
    )
    if self.max_categorical_size > 0:
      features_categorical_logits = (
          self._create_categorical_feature_logits(
              features.categorical, features_batch.categorical, scale
          )
          + perturbations_batch.categorical
      )
      new_features_categorical = tfd.Categorical(
          logits=features_categorical_logits
      ).sample(seed=categorical_seed)
    else:
      new_features_categorical = jnp.zeros(
          features_batch.continuous.shape[:2] + (0,), dtype=types.INT_DTYPE
      )
    return vb.VectorizedOptimizerInput(
        continuous=new_features_continuous, categorical=new_features_categorical
    )

  def _create_logits_vector(
      self,
      features_one_category: jax.Array,  # [pool_size]
      feature_batch_member_one_category: jax.Array,  # scalar integer
      scale_batch_member: jax.Array,  # [pool_size]
      feature_size: jax.Array,
  ):  # scalar
    categories = jnp.arange(self.max_categorical_size)
    logit_same_category = jnp.log(
        self.config.prob_same_category_without_perturbation
    )
    logit_different_category = jnp.log(
        (1.0 - self.config.prob_same_category_without_perturbation)
        / (feature_size - 1.0)
    )
    logits = (
        jnp.sum(
            jnp.where(
                categories[:, jnp.newaxis] == features_one_category,
                scale_batch_member,
                0.0,
            ),
            axis=-1,
        )
        + logit_different_category
    )
    logits = jnp.where(categories < feature_size, logits, -jnp.inf)
    return logits.at[feature_batch_member_one_category].add(
        -jnp.sum(scale_batch_member)
        + logit_same_category
        - logit_different_category
    )  # [num_categories]

  def _create_logits_one_feature(
      self,
      features_one_category: jax.Array,  # [pool_size, num_parallel]
      features_batch_one_category: jax.Array,  # [batch_size, num_parallel]
      scale: jax.Array,  # [batch_size, pool_size]
      feature_size: jax.Array,  # scalar
  ):
    return jax.vmap(  # map over batch
        jax.vmap(self._create_logits_vector, in_axes=(-1, -1, None, None)),
        in_axes=(None, 0, 0, None),
    )(
        features_one_category, features_batch_one_category, scale, feature_size
    )  # [batch_size, num_parallel, max_num_categories]

  def _create_categorical_feature_logits(
      self,
      features: jax.Array,  # [pool_size, num_parallel, num_features]
      features_batch: jax.Array,  # [batch_size, num_parallel, num_features]
      scale: jax.Array,  # [batch_size, pool_size]
  ):
    return jax.vmap(
        self._create_logits_one_feature, in_axes=(-1, -1, None, 0), out_axes=2
    )(
        features, features_batch, scale, jnp.array(self.categorical_sizes)
    )  # [batch_size, num_parallel, num_features, num_categories]

  def _create_random_perturbations(
      self,
      perturbations_batch: jax.Array,
      n_parallel: int,
      seed: jax.Array,
  ) -> types.ContinuousAndCategoricalArray:
    """Create random perturbations for the newly created batch.

    Args:
      perturbations_batch: (batch_size,)
      n_parallel: Number of points that the acquisition function maps to a
        single value. This arg may be greater than 1 if a parallel acquisition
        function (qEI, qUCB) is used; otherwise it should be 1.
      seed: Random seed.

    Returns:
      perturbations: (batch_size, n_parallel, n_features)
    """
    cont_seed, cat_seed = jax.random.split(seed)
    # Generate normalized noise for each batch.
    batch_noise_continuous = jax.random.laplace(
        cont_seed,
        shape=(
            self.batch_size,
            n_parallel,
            self.n_feature_dimensions_with_padding.continuous,
        ),
    )
    if self.n_feature_dimensions_with_padding.continuous > 0:
      batch_noise_continuous /= jnp.max(
          jnp.abs(batch_noise_continuous), axis=1, keepdims=True
      )

    if self.n_feature_dimensions_with_padding.continuous == 0:
      categorical_perturbation = (
          self.config.pure_categorical_perturbation_factor
      )
    else:
      categorical_perturbation = self.config.categorical_perturbation_factor
    batch_noise_categorical = (
        jax.random.laplace(
            cat_seed,
            shape=(
                self.batch_size,
                n_parallel,
                self.n_feature_dimensions_with_padding.categorical,
                self.max_categorical_size,
            ),
        )
        * categorical_perturbation
    )
    return types.ContinuousAndCategoricalArray(
        continuous=(
            batch_noise_continuous
            * perturbations_batch[:, jnp.newaxis, jnp.newaxis]
        ),
        categorical=(
            batch_noise_categorical
            * perturbations_batch[:, jnp.newaxis, jnp.newaxis, jnp.newaxis]
        ),
    )

  def update(
      self,
      seed: jax.Array,
      state: VectorizedEagleStrategyState,
      batch_features: vb.VectorizedOptimizerInput,
      batch_rewards: types.Array,
  ) -> VectorizedEagleStrategyState:
    """Update the firefly pool based on the new batch of results.

    Arguments:
      seed: Random seed.
      state: Current state.
      batch_features: (batch_size, n_parallel, n_features)
      batch_rewards: (batch_size, )

    Returns:
      new_state: Updated state.
    """
    new_best_reward = jnp.maximum(state.best_reward, jnp.max(batch_rewards))
    batch_id = state.iterations % (self.pool_size // self.batch_size)
    batch_start_ind = batch_id * self.batch_size
    batch_perturbations = jax.lax.dynamic_slice_in_dim(
        state.perturbations, batch_start_ind, self.batch_size
    )

    def _update(batch_features, batch_rewards, batch_perturbations):
      # Pass the new batch rewards and the associated last suggested features.
      new_batch_features, new_batch_rewards, new_batch_perturbations = (
          self._update_pool_features_and_rewards(
              batch_features,
              batch_rewards,
              jax.tree_util.tree_map(
                  lambda f: jax.lax.dynamic_slice_in_dim(
                      f, batch_start_ind, self.batch_size
                  ),
                  state.features,
              ),
              jax.lax.dynamic_slice_in_dim(
                  state.rewards, batch_start_ind, self.batch_size
              ),
              batch_perturbations,
          )
      )
      return self._trim_pool(
          new_batch_features,
          new_batch_rewards,
          new_batch_perturbations,
          new_best_reward,
          seed,
      )

    # If the strategy is still initializing, return the random/prior values.
    (new_batch_features, new_batch_rewards, new_batch_perturbations) = (
        jax.lax.cond(
            state.iterations < self.pool_size // self.batch_size,
            lambda *args: args,
            _update,
            batch_features,
            batch_rewards,
            batch_perturbations,
        )
    )

    return VectorizedEagleStrategyState(
        iterations=state.iterations + 1,
        features=jax.tree_util.tree_map(
            lambda sf, nbf: jax.lax.dynamic_update_slice_in_dim(
                sf, nbf, batch_start_ind, axis=0
            ),
            state.features,
            new_batch_features,
        ),
        rewards=jax.lax.dynamic_update_slice_in_dim(
            state.rewards, new_batch_rewards, batch_start_ind, axis=0
        ),
        best_reward=new_best_reward,
        perturbations=jax.lax.dynamic_update_slice_in_dim(
            state.perturbations,
            new_batch_perturbations,
            batch_start_ind,
            axis=0,
        ),
    )

  def _update_pool_features_and_rewards(
      self,
      batch_features: vb.VectorizedOptimizerInput,
      batch_rewards: jax.Array,
      prev_batch_features: vb.VectorizedOptimizerInput,
      prev_batch_rewards: jax.Array,
      perturbations: jax.Array,
  ) -> Tuple[vb.VectorizedOptimizerInput, jax.Array, jax.Array]:
    """Update the features and rewards for fireflies with improved rewards.

    Arguments:
      batch_features: (batch_size, n_parallel, n_features), new proposed
        features batch.
      batch_rewards: (batch_size,), rewards for new proposed features batch.
      prev_batch_features: (batch_size, n_parallel, n_features), previous
        features batch.
      prev_batch_rewards: (batch_size,), rewards for previous features batch.
      perturbations: (batch_size,)

    Returns:
      sliced features, sliced rewards, sliced perturbations
    """
    # Find indices of flies that their generated features made an improvement.
    improve_indx = batch_rewards > prev_batch_rewards
    # Update successful flies' with the associated last features and rewards.
    new_batch_features = jax.tree_util.tree_map(
        lambda bf, pbf: jnp.where(
            improve_indx[..., jnp.newaxis, jnp.newaxis], bf, pbf
        ),
        batch_features,
        prev_batch_features,
    )
    new_batch_rewards = jnp.where(
        improve_indx, batch_rewards, prev_batch_rewards
    )
    # Penalize unsuccessful flies.
    new_batch_perturbations = jnp.where(
        improve_indx, perturbations, perturbations * self.config.penalize_factor
    )
    return new_batch_features, new_batch_rewards, new_batch_perturbations

  def _trim_pool(
      self,
      batch_features: vb.VectorizedOptimizerInput,
      batch_rewards: jax.Array,
      batch_perturbations: jax.Array,
      best_reward: jax.Array,
      seed: jax.Array,
  ) -> Tuple[vb.VectorizedOptimizerInput, jax.Array, jax.Array]:
    """Trim the pool by replacing unsuccessful fireflies with new random ones.

    A firefly is considered unsuccessful if its current perturbation is below
    'perturbation_lower_bound' and it's not the best fly seen thus far.
    Random features are created to replace the existing ones, and rewards
    are set to -np.inf to indicate that we don't yet have values for those
    features and they shouldn't be used during suggest.

    Args:
      batch_features: (batch_size, n_parallel, n_features)
      batch_rewards: (batch_size,)
      batch_perturbations: (batch_size,)
      best_reward: Best reward seen so far.
      seed: Random seed.

    Returns:
      updated feature, reward, and perturbation batches.
    """
    indx = batch_perturbations < self.config.perturbation_lower_bound
    # Ensure the best firefly is never removed. For optimization purposes,
    # this logic is inside the if statement to be peformed only if needed.
    indx = indx & (batch_rewards != best_reward)

    # Replace fireflies with random features and evaluate rewards.
    random_features = self.random_sampler(
        self.batch_size,
        n_parallel=batch_features.continuous.shape[1],
        seed=seed,
    )
    new_batch_features = jax.tree_util.tree_map(
        lambda rf, bf: jnp.where(indx[..., jnp.newaxis, jnp.newaxis], rf, bf),
        random_features,
        batch_features,
    )
    new_batch_perturbations = jnp.where(
        indx, self.config.perturbation, batch_perturbations
    )
    # Setting rewards to -inf to filter out those fireflies during suggest.
    new_batch_rewards = jnp.where(indx, -jnp.inf, batch_rewards)
    return new_batch_features, new_batch_rewards, new_batch_perturbations



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for meta_learning_utils."""

import random
from vizier import pyvizier as vz
from vizier._src.algorithms.designers.meta_learning import meta_learning_utils
from absl.testing import absltest
from absl.testing import parameterized


class MetaLearningUtilsTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    space = vz.SearchSpace()
    space.root.add_int_param('tuned_param', 0, 100, default_value=55)
    self.utils = meta_learning_utils.MetaLearningUtils(
        goal=vz.ObjectiveMetricGoal.MAXIMIZE,
        tuned_metric_name='tuned_obj',
        meta_metric_name='meta_obj',
        tuning_params=space,
    )
    self.meta_trials = []
    for i in range(10):
      meta_trial = vz.Trial({'meta_param': i})
      meta_trial.complete(vz.Measurement(metrics={'meta_obj': float(i)}))
      self.meta_trials.append(meta_trial)
    random.shuffle(self.meta_trials)

    self.tuned_trials = []
    for i in range(50):
      tuned_trial = vz.Trial({'tuned_param': i})
      tuned_trial.complete(vz.Measurement(metrics={'tuned_obj': float(i)}))
      self.tuned_trials.append(tuned_trial)
    random.shuffle(self.tuned_trials)

  def test_best_trial(self):
    # meta trial
    best_meta_trial = self.utils.get_best_meta_trial(self.meta_trials)
    self.assertEqual(best_meta_trial.parameters['meta_param'].value, 9)
    # tuned trial
    best_tuned_trial = self.utils.get_best_tuned_trial(self.tuned_trials)
    self.assertEqual(best_tuned_trial.parameters['tuned_param'].value, 49)

  def test_best_trial_score(self):
    # meta trial score
    best_meta_trial_score = self.utils.get_best_meta_trial_score(
        self.meta_trials
    )
    self.assertEqual(best_meta_trial_score, 9.0)
    # tuned trial score
    best_tuned_trial_score = self.utils.get_best_tuned_trial_score(
        self.tuned_trials
    )
    self.assertEqual(best_tuned_trial_score, 49.0)

  def test_generate_default_tuned_parameters(self):
    tuned_suggestion = self.utils.get_default_hyperparameters()
    self.assertEqual(tuned_suggestion.parameters['tuned_param'].value, 55)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for singleton_params module which allows to handle singleton parameters."""

from vizier import pyvizier as vz
from vizier._src.pythia import singleton_params
from absl.testing import absltest
from absl.testing import parameterized


def get_problem_with_singletons() -> vz.ProblemStatement:
  """Returns a problem with singleton parameters."""
  problem = vz.ProblemStatement()
  problem.search_space.root.add_float_param('sf', min_value=1, max_value=1)
  problem.search_space.root.add_int_param('si', min_value=5, max_value=5)
  problem.search_space.root.add_categorical_param('sc', feasible_values=['a'])
  problem.search_space.root.add_discrete_param('sd', feasible_values=[3])

  problem.search_space.root.add_float_param('f', min_value=0.0, max_value=5.0)
  problem.search_space.root.add_int_param('i', min_value=0, max_value=10)
  problem.search_space.root.add_categorical_param(
      'c', feasible_values=['a', '1']
  )
  problem.search_space.root.add_discrete_param('d', feasible_values=[3, 1])
  problem.metric_information = [
      vz.MetricInformation(name='obj', goal=vz.ObjectiveMetricGoal.MAXIMIZE)
  ]
  return problem


def get_problem_without_singletons() -> vz.ProblemStatement:
  """Returns a problem with singleton parameters."""
  problem = vz.ProblemStatement()
  problem.search_space.root.add_float_param('f', min_value=0.0, max_value=5.0)
  problem.search_space.root.add_int_param('i', min_value=0, max_value=10)
  problem.search_space.root.add_categorical_param(
      'c', feasible_values=['a', '1']
  )
  problem.search_space.root.add_discrete_param('d', feasible_values=[3, 1])
  problem.metric_information = [
      vz.MetricInformation(name='obj', goal=vz.ObjectiveMetricGoal.MAXIMIZE)
  ]
  return problem


class SingletonParameterHandlerTest(parameterized.TestCase):
  """Tests for the `singletonParameterHandler`."""

  @parameterized.named_parameters(
      ('with_singletons', get_problem_with_singletons()),
      ('without_singletons', get_problem_without_singletons()),
  )
  def test_stripped_problem(self, problem):
    handler = singleton_params.SingletonParameterHandler(problem)
    expected_search_search = get_problem_without_singletons().search_space
    self.assertEqual(
        handler.stripped_problem.search_space, expected_search_search
    )

  def test_strip_trials(self):
    problem = get_problem_with_singletons()
    handler = singleton_params.SingletonParameterHandler(problem)
    trials = [
        vz.TrialSuggestion(
            parameters={
                'sf': 1.0,
                'si': 5,
                'sc': 'a',
                'sd': 3,
                'f': 1.0,
                'i': 1,
                'c': 'a',
                'd': 3,
            }
        )
    ]
    stripped_trials = handler.strip_trials(trials)
    self.assertEqual(
        stripped_trials,
        [
            vz.TrialSuggestion(
                parameters={
                    'f': 1.0,
                    'i': 1,
                    'c': 'a',
                    'd': 3,
                }
            )
        ],
    )

  def test_augment_trials(self):
    problem = get_problem_with_singletons()
    handler = singleton_params.SingletonParameterHandler(problem)
    trials = [
        vz.TrialSuggestion(
            parameters={
                'f': 1.0,
                'i': 1,
                'c': 'a',
                'd': 3,
            }
        )
    ]
    augmented_trials = handler.augment_trials(trials)
    self.assertEqual(
        augmented_trials,
        [
            vz.TrialSuggestion(
                parameters={
                    'sf': 1.0,
                    'si': 5,
                    'sc': 'a',
                    'sd': 3,
                    'f': 1.0,
                    'i': 1,
                    'c': 'a',
                    'd': 3,
                }
            )
        ],
    )

  def test_find_singletons(self):
    problem = get_problem_with_singletons()
    handler = singleton_params.SingletonParameterHandler(problem)
    self.assertEqual(
        handler._singletons,
        {
            'sf': 1.0,
            'si': 5,
            'sc': 'a',
            'sd': 3,
        },
    )

  def test_apply_strip_multiple_times(self):
    problem = get_problem_with_singletons()
    handler = singleton_params.SingletonParameterHandler(problem)
    trials = [
        vz.TrialSuggestion(
            parameters={
                'sf': 1.0,
                'si': 5,
                'sc': 'a',
                'sd': 3,
                'f': 1.0,
                'i': 1,
                'c': 'a',
                'd': 3,
            }
        )
    ]
    for _ in range(5):
      trials = handler.strip_trials(trials)
    self.assertEqual(
        trials,
        [
            vz.TrialSuggestion(
                parameters={
                    'f': 1.0,
                    'i': 1,
                    'c': 'a',
                    'd': 3,
                }
            )
        ],
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from jax import numpy as jnp
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import scalarization
from vizier._src.algorithms.designers import scalarizing_designer
from vizier._src.algorithms.designers.eagle_strategy import eagle_strategy
from vizier._src.algorithms.testing import test_runners
from vizier.testing import test_studies

from absl.testing import absltest


class ScalarizingDesignerTest(absltest.TestCase):

  def test_scalarizing_eagle(self):
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.extend([
        vz.MetricInformation(
            name='metric1', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        ),
        vz.MetricInformation(
            name='metric2', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        ),
    ])

    def eagle_designer_factory(ps, seed):
      return eagle_strategy.EagleStrategyDesigner(
          problem_statement=ps, seed=seed
      )

    scalarized_designer = scalarizing_designer.ScalarizingDesigner(
        problem,
        eagle_designer_factory,
        scalarizer=scalarization.HyperVolumeScalarization(
            weights=jnp.ones(len(problem.metric_information))
        ),
    )
    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=3,
            batch_size=5,
            verbose=1,
            validate_parameters=True,
        ).run_designer(scalarized_designer),
        15,
    )

  def test_ensemble_scalarizing_eagle(self):
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.extend([
        vz.MetricInformation(
            name='metric1', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        ),
        vz.MetricInformation(
            name='metric2', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        ),
    ])

    def eagle_designer_factory(ps, seed):
      return eagle_strategy.EagleStrategyDesigner(
          problem_statement=ps, seed=seed
      )

    scalarized_designer = (
        scalarizing_designer.create_gaussian_scalarizing_designer(
            problem,
            eagle_designer_factory,
            scalarization.HyperVolumeScalarization,
            num_ensemble=10,
        )
    )
    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=3,
            batch_size=5,
            verbose=1,
            validate_parameters=True,
        ).run_designer(scalarized_designer),
        15,
    )

  def test_missing_metrics(self):
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.extend([
        vz.MetricInformation(
            name='metric1', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        ),
        vz.MetricInformation(
            name='metric2', goal=vz.ObjectiveMetricGoal.MAXIMIZE
        ),
    ])

    def eagle_designer_factory(ps, seed):
      return eagle_strategy.EagleStrategyDesigner(
          problem_statement=ps, seed=seed
      )

    scalarized_designer = scalarizing_designer.ScalarizingDesigner(
        problem,
        eagle_designer_factory,
        scalarizer=scalarization.HyperVolumeScalarization(
            weights=jnp.ones(len(problem.metric_information))
        ),
    )

    suggestions = scalarized_designer.suggest(1)
    trial = list(suggestions)[0].to_trial(1)
    trial.complete(vz.Measurement(metrics={'metric1': 0.4}))
    scalarized_designer.update(
        vza.CompletedTrials(trials=[trial]),
        all_active=vza.ActiveTrials(trials=[]),
    )
    self.assertTrue(
        jnp.isnan(trial.final_measurement_or_die.metrics['scalarized'].value)
    )


if __name__ == '__main__':
  absltest.main()
